import numba as nb
import numpy as np

__all__ = ['LU_Biharmonic', 'Biharmonic_factor_pr', 'Biharmonic_Solve',
           'Biharmonic_matvec']

def LU_Biharmonic(a0, alfa, beta, sii, siu, siuu, ail, aii, aiu,
                  bill, bil, bii, biu, biuu, u0, u1,
                  u2, l0, l1, axis):
    if l1.ndim == 2:
        LU_Biharmonic_1D(a0, np.atleast_1d(alfa).item(), np.atleast_1d(beta).item(),
                         sii, siu, siuu, ail, aii, aiu,
                         bill, bil, bii, biu, biuu, u0, u1,
                         u2, l0, l1)
    elif l1.ndim == 3:
        LU_Biharmonic_2D(a0, alfa, beta, sii, siu, siuu, ail, aii, aiu,
                         bill, bil, bii, biu, biuu, u0, u1,
                         u2, l0, l1, axis)
    elif l1.ndim == 4:
        LU_Biharmonic_3D(a0, alfa, beta, sii, siu, siuu, ail, aii, aiu,
                         bill, bil, bii, biu, biuu, u0, u1,
                         u2, l0, l1, axis)

@nb.jit(nopython=True, fastmath=True, cache=True)
def LU_Biharmonic_2D(a0, alfa, beta, sii, siu, siuu, ail, aii, aiu,
                     bill, bil, bii, biu, biuu, u0, u1,
                     u2, l0, l1, axis):
    if axis == 0:
        for j in range(l1.shape[2]):
            LU_Biharmonic_1D(a0, alfa[0, j], beta[0, j],
                             sii, siu, siuu, ail, aii, aiu,
                             bill, bil, bii, biu, biuu,
                             u0[:, :, j], u1[:, :, j],
                             u2[:, :, j], l0[:, :, j], l1[:, :, j])
    elif axis == 1:
        for i in range(l1.shape[1]):
            LU_Biharmonic_1D(a0, alfa[i, 0], beta[i, 0],
                             sii, siu, siuu, ail, aii, aiu,
                             bill, bil, bii, biu, biuu,
                             u0[:, i], u1[:, i],
                             u2[:, i], l0[:, i], l1[:, i])

@nb.jit(nopython=True, fastmath=True, cache=True)
def LU_Biharmonic_3D(a0, alfa, beta, sii, siu, siuu, ail, aii, aiu,
                     bill, bil, bii, biu, biuu, u0, u1,
                     u2, l0, l1, axis):
    if axis == 0:
        for j in range(l1.shape[2]):
            for k in range(l1.shape[3]):
                LU_Biharmonic_1D(a0, alfa[0, j, k], beta[0, j, k],
                                 sii, siu, siuu, ail, aii, aiu,
                                 bill, bil, bii, biu, biuu,
                                 u0[:, :, j, k], u1[:, :, j, k],
                                 u2[:, :, j, k], l0[:, :, j, k], l1[:, :, j, k])
    elif axis == 1:
        for i in range(l1.shape[1]):
            for k in range(l1.shape[3]):
                LU_Biharmonic_1D(a0, alfa[i, 0, k], beta[i, 0, k],
                                 sii, siu, siuu, ail, aii, aiu,
                                 bill, bil, bii, biu, biuu,
                                 u0[:, i, :, k], u1[:, i, :, k],
                                 u2[:, i, :, k], l0[:, i, :, k], l1[:, i, :, k])
    elif axis == 2:
        for i in range(l1.shape[1]):
            for j in range(l1.shape[2]):
                LU_Biharmonic_1D(a0, alfa[i, j, 0], beta[i, j, 0],
                                 sii, siu, siuu, ail, aii, aiu,
                                 bill, bil, bii, biu, biuu,
                                 u0[:, i, j], u1[:, i, j],
                                 u2[:, i, j], l0[:, i, j], l1[:, i, j])

@nb.jit(nopython=True, fastmath=True, cache=True)
def LU_Biharmonic_1D(a, b, c,
                     # 3 upper diagonals of SBB
                     sii, siu, siuu,
                     # All 3 diagonals of ABB
                     ail, aii, aiu,
                     # All 5 diagonals of BBB
                     bill, bil, bii, biu, biuu,
                     # Three upper and two lower diagonals of LU decomposition
                     u0, u1, u2, l0, l1):
    LU_oe_Biharmonic_1D(0, a, b, c, sii[::2], siu[::2], siuu[::2], ail[::2],
                        aii[::2], aiu[::2], bill[::2], bil[::2], bii[::2],
                        biu[::2], biuu[::2], u0[0], u1[0], u2[0], l0[0], l1[0])
    LU_oe_Biharmonic_1D(1, a, b, c, sii[1::2], siu[1::2], siuu[1::2], ail[1::2],
                        aii[1::2], aiu[1::2], bill[1::2], bil[1::2], bii[1::2],
                        biu[1::2], biuu[1::2], u0[1], u1[1], u2[1], l0[1], l1[1])

@nb.jit(nopython=True, fastmath=True, cache=True)
def LU_oe_Biharmonic_1D(odd, a, b, c,
                        # 3 upper diagonals of SBB
                        sii, siu, siuu,
                        # All 3 diagonals of ABB
                        ail, aii, aiu,
                        # All 5 diagonals of BBB
                        bill, bil, bii, biu, biuu,
                        # Two upper and two lower diagonals of LU decomposition
                        u0, u1, u2, l0, l1):

    M = sii.shape[0]
    c0 = np.zeros(M)
    c1 = np.zeros(M)
    c2 = np.zeros(M)

    c0[0] = a*sii[0] + b*aii[0] + c*bii[0]
    c0[1] = a*siu[0] + b*aiu[0] + c*biu[0]
    c0[2] = a*siuu[0] + c*biuu[0]
    m = 8*(odd+1)*(odd+2)*(odd*(odd+4)+3*(6+odd+2)*(6+odd+2))
    c0[3] = m*a*np.pi/(6+odd+3.)
    #c0[3] = a*8./(6+odd+3.)*pi*(odd+1.)*(odd+2.)*(odd*(odd+4.)+3.*pow(6+odd+2., 2))
    m = 8*(odd+1)*(odd+2)*(odd*(odd+4)+3*(8+odd+2)*(8+odd+2))
    c0[4] = m*a*np.pi/(8+odd+3.)
    #c0[4] = a*8./(8+odd+3.)*pi*(odd+1.)*(odd+2.)*(odd*(odd+4.)+3.*pow(8+odd+2., 2))
    c1[0] = b*ail[0] + c*bil[0]
    c1[1] = a*sii[1] + b*aii[1] + c*bii[1]
    c1[2] = a*siu[1] + b*aiu[1] + c*biu[1]
    c1[3] = a*siuu[1] + c*biuu[1]
    m = 8*(odd+3)*(odd+4)*((odd+2)*(odd+6)+3*(8+odd+2)*(8+odd+2))
    c1[4] = m*a*np.pi/(8+odd+3.)
    #c1[4] = a*8./(8+odd+3.)*pi*(odd+3.)*(odd+4.)*((odd+2.)*(odd+6.)+3.*pow(8+odd+2., 2))
    c2[0] = c*bill[0]
    c2[1] = b*ail[1] + c*bil[1]
    c2[2] = a*sii[2] + b*aii[2] + c*bii[2]
    c2[3] = a*siu[2] + b*aiu[2] + c*biu[2]
    c2[4] = a*siuu[2] + c*biuu[2]
    for i in range(5, M):
        j = 2*i+odd
        m = 8*(odd+1)*(odd+2)*(odd*(odd+4)+3*(j+2)*(j+2))
        c0[i] = m*a*np.pi/(j+3.)
        m = 8*(odd+3)*(odd+4)*((odd+2)*(odd+6)+3*(j+2)*(j+2))
        c1[i] = m*a*np.pi/(j+3.)
        m = 8*(odd+5)*(odd+6)*((odd+4)*(odd+8)+3*(j+2)*(j+2))
        c2[i] = m*a*np.pi/(j+3.)
        #c0[i] = a*8./(j+3.)*pi*(odd+1.)*(odd+2.)*(odd*(odd+4.)+3.*pow(j+2., 2))
        #c1[i] = a*8./(j+3.)*pi*(odd+3.)*(odd+4.)*((odd+2)*(odd+6.)+3.*pow(j+2., 2))
        #c2[i] = a*8./(j+3.)*pi*(odd+5.)*(odd+6.)*((odd+4)*(odd+8.)+3.*pow(j+2., 2))

    u0[0] = c0[0]
    u1[0] = c0[1]
    u2[0] = c0[2]
    for kk in range(1, M):
        l0[kk-1] = c1[kk-1]/u0[kk-1]
        if kk < M-1:
            l1[kk-1] = c2[kk-1]/u0[kk-1]

        for i in range(kk, M):
            c1[i] = c1[i] - l0[kk-1]*c0[i]

        if kk < M-1:
            for i in range(kk, M):
                c2[i] = c2[i] - l1[kk-1]*c0[i]

        for i in range(kk, M):
            c0[i] = c1[i]
            c1[i] = c2[i]

        if kk < M-2:
            c2[kk] = c*bill[kk]
            c2[kk+1] = b*ail[kk+1] + c*bil[kk+1]
            c2[kk+2] = a*sii[kk+2] + b*aii[kk+2] + c*bii[kk+2]
            if kk < M-3:
                c2[kk+3] = a*siu[kk+2] + b*aiu[kk+2] + c*biu[kk+2]
            if kk < M-4:
                c2[kk+4] = a*siuu[kk+2] + c*biuu[kk+2]
            if kk < M-5:
                k = 2*(kk+2)+odd
                for i in range(kk+5, M):
                    j = 2*i+odd
                    m = 8*(k+1)*(k+2)*(k*(k+4)+3*(j+2)*(j+2))
                    c2[i] = m*a*np.pi/(j+3.)
                    #c2[i] = a*8./(j+3.)*pi*(k+1.)*(k+2.)*(k*(k+4.)+3.*pow(j+2., 2))

        u0[kk] = c0[kk]
        if kk < M-1:
            u1[kk] = c0[kk+1]
        if kk < M-2:
            u2[kk] = c0[kk+2]

@nb.jit(nopython=True, fastmath=True, cache=True)
def Biharmonic_factor_pr_3D(axis, a, b, l0, l1):
    if axis == 0:
        for ii in range(a.shape[2]):
            for jj in range(a.shape[3]):
                Biharmonic_factor_pr_1D(a[:, :, ii, jj],
                                        b[:, :, ii, jj],
                                        l0[:, :, ii, jj],
                                        l1[:, :, ii, jj])
    elif axis == 1:
        for ii in range(a.shape[1]):
            for jj in range(a.shape[3]):
                Biharmonic_factor_pr_1D(a[:, ii, :, jj],
                                        b[:, ii, :, jj],
                                        l0[:, ii, :, jj],
                                        l1[:, ii, :, jj])

    elif axis == 2:
        for ii in range(a.shape[1]):
            for jj in range(a.shape[2]):
                Biharmonic_factor_pr_1D(a[:, ii, jj, :],
                                        b[:, ii, jj, :],
                                        l0[:, ii, jj, :],
                                        l1[:, ii, jj, :])

@nb.jit(nopython=True, fastmath=True, cache=True)
def Biharmonic_factor_pr_2D(axis, a, b, l0, l1):
    if axis == 0:
        for ii in range(a.shape[2]):
            Biharmonic_factor_pr_1D(a[:, :, ii],
                                    b[:, :, ii],
                                    l0[:, :, ii],
                                    l1[:, :, ii])
    elif axis == 1:
        for ii in range(a.shape[1]):
            Biharmonic_factor_pr_1D(a[:, ii, :],
                                    b[:, ii, :],
                                    l0[:, ii, :],
                                    l1[:, ii, :])

@nb.jit(nopython=True, fastmath=True, cache=True)
def Biharmonic_factor_pr_1D(a, b, l0, l1):
    Biharmonic_factor_oe_pr(0, a[0], b[0], l0[0], l1[0])
    Biharmonic_factor_oe_pr(1, a[1], b[1], l0[1], l1[1])

@nb.jit(nopython=True, fastmath=True, cache=True)
def Biharmonic_factor_oe_pr(odd, a, b, l0, l1):
    M = l0.shape[0]+1
    k = odd
    a[0] = 8*k*(k+1)*(k+2)*(k+4)*np.pi
    b[0] = 24*(k+1)*(k+2)*np.pi
    k = 2+odd
    a[1] = 8*k*(k+1)*(k+2)*(k+4)*np.pi - l0[0]*a[0]
    b[1] = 24*(k+1)*(k+2)*np.pi - l0[0]*b[0]
    for k in range(2, M-3):
        kk = 2*k+odd
        pp = 8*kk*(kk+1)*(kk+2)*(kk+4)
        rr = 24*(kk+1)*(kk+2)
        a[k] = pp*np.pi - l0[k-1]*a[k-1] - l1[k-2]*a[k-2]
        b[k] = rr*np.pi - l0[k-1]*b[k-1] - l1[k-2]*b[k-2]

def Biharmonic_factor_pr(a, b, l0, l1, axis):
    if a.ndim == 2:
        Biharmonic_factor_pr_1D(a, b, l0, l1)
    elif a.ndim == 3:
        Biharmonic_factor_pr_2D(axis, a, b, l0, l1)
    elif a.ndim == 4:
        Biharmonic_factor_pr_3D(axis, a, b, l0, l1)

def Biharmonic_Solve(b, u, u0, u1, u2, l0, l1, ak, bk, a0, axis=0):
    if b.ndim == 1:
        Solve_Biharmonic_1D(b, u, u0, u1, u2, l0, l1, ak, bk, a0)
    elif b.ndim == 2:
        Solve_Biharmonic_2D(axis, b, u, u0, u1, u2, l0, l1, ak, bk, a0)
    elif b.ndim == 3:
        Solve_Biharmonic_3D(axis, b, u, u0, u1, u2, l0, l1, ak, bk, a0)

@nb.jit(nopython=True, fastmath=True, cache=True)
def Solve_Biharmonic_1D(fk, uk, u0, u1, u2, l0, l1, a, b, ac):
    Solve_oe_Biharmonic_1D(0, fk[::2], uk[::2], u0[0], u1[0], u2[0], l0[0], l1[0], a[0], b[0], ac)
    Solve_oe_Biharmonic_1D(1, fk[1::2], uk[1::2], u0[1], u1[1], u2[1], l0[1], l1[1], a[1], b[1], ac)

@nb.jit(nopython=True, fastmath=True, cache=True)
def BackBsolve_U(M, odd, f, uk, u0, u1, u2, l0, l1, a, b, ac):
    s1 = 0.0
    s2 = 0.0

    uk[M-1] = f[M-1] / u0[M-1]
    uk[M-2] = (f[M-2] - u1[M-2]*uk[M-1]) / u0[M-2]
    uk[M-3] = (f[M-3] - u1[M-3]*uk[M-2] - u2[M-3]*uk[M-1]) / u0[M-3]

    for kk in range(M-4, -1, -1):
        k = 2*kk+odd
        j = k+6
        s1 += uk[kk+3]/(j+3.)
        s2 += (uk[kk+3]/(j+3.))*((j+2)*(j+2))
        uk[kk] = (f[kk] - u1[kk]*uk[kk+1] - u2[kk]*uk[kk+2] - a[kk]*ac*s1 - b[kk]*ac*s2) / u0[kk]

@nb.jit(nopython=True, fastmath=True, cache=True)
def Solve_oe_Biharmonic_1D(odd, fk, uk, u0, u1, u2, l0, l1, a, b, ac):
    """
    Solve (aS+b*A+cB)x = f, where S, A and B are 4th order Laplace, stiffness and mass matrices of Shen with Dirichlet BC
    """
    y = np.zeros(u0.shape[0], dtype=fk.dtype)
    M = u0.shape[0]
    ForwardBsolve_L(y, l0, l1, fk)

    # Solve Backward U u = y
    BackBsolve_U(M, odd, y, uk, u0, u1, u2, l0, l1, a, b, ac)

@nb.jit(nopython=True, fastmath=True, cache=True)
def ForwardBsolve_L(y, l0, l1, fk):
    # Solve Forward Ly = f
    y[0] = fk[0]
    y[1] = fk[1] - l0[0]*y[0]
    N = l0.shape[0]
    for i in range(2, N):
        y[i] = fk[i] - l0[i-1]*y[i-1] - l1[i-2]*y[i-2]

@nb.jit(nopython=True, fastmath=True, cache=True)
def Solve_Biharmonic_2D(axis, b, u, u0, u1, u2, l0, l1, ak, bk, a0):
    if axis == 0:
        for j in range(b.shape[1]):
            Solve_Biharmonic_1D(b[:, j], u[:, j], u0[:, :, j], u1[:, :, j],
                                u2[:, :, j], l0[:, :, j], l1[:, :, j],
                                ak[:, :, j], bk[:, :, j], a0)
    elif axis == 1:
        for i in range(b.shape[0]):
            Solve_Biharmonic_1D(b[i], u[i], u0[:, i], u1[:, i],
                                u2[:, i], l0[:, i], l1[:, i],
                                ak[:, i], bk[:, i], a0)

@nb.jit(nopython=True, fastmath=True, cache=True)
def Solve_Biharmonic_3D(axis, b, u, u0, u1, u2, l0, l1, ak, bk, a0):
    if axis == 0:
        for j in range(b.shape[1]):
            for k in range(b.shape[2]):
                Solve_Biharmonic_1D(b[:, j, k], u[:, j, k], u0[:, :, j, k], u1[:, :, j, k],
                                    u2[:, :, j, k], l0[:, :, j, k], l1[:, :, j, k],
                                    ak[:, :, j, k], bk[:, :, j, k], a0)
    elif axis == 1:
        for i in range(b.shape[0]):
            for k in range(b.shape[2]):
                Solve_Biharmonic_1D(b[i, :, k], u[i, :, k], u0[:, i, :, k], u1[:, i, :, k],
                                    u2[:, i, :, k], l0[:, i, :, k], l1[:, i, :, k],
                                    ak[:, i, :, k], bk[:, i, :, k], a0)
    elif axis == 2:
        for i in range(b.shape[0]):
            for j in range(b.shape[1]):
                Solve_Biharmonic_1D(b[i, j], u[i, j], u0[:, i, j], u1[:, i, j],
                                    u2[:, i, j], l0[:, i, j], l1[:, i, j],
                                    ak[:, i, j], bk[:, i, j], a0)

@nb.jit(nopython=True, fastmath=True, cache=True)
def Biharmonic_matvec2D(v, b, a0, alfa, beta,
                        sii, siu, siuu,
                        ail, aii, aiu,
                        bill, bil, bii, biu, biuu, axis):
    if axis == 0:
        for j in range(v.shape[1]):
            Biharmonic_matvec1D(v[:, j], b[:, j], a0, alfa[0, j],
                                beta[0, j], sii, siu, siuu,
                                ail, aii, aiu,
                                bill, bil, bii, biu, biuu)
    elif axis == 1:
        for i in range(v.shape[0]):
            Biharmonic_matvec1D(v[i], b[i], a0, alfa[i, 0],
                                beta[i, 0], sii, siu, siuu,
                                ail, aii, aiu,
                                bill, bil, bii, biu, biuu)

@nb.jit(nopython=True, fastmath=True, cache=True)
def Biharmonic_matvec3D(v, b, a0, alfa, beta,
                        sii, siu, siuu,
                        ail, aii, aiu,
                        bill, bil, bii, biu, biuu, axis):
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                Biharmonic_matvec1D(v[:, j, k], b[:, j, k], a0, alfa[0, j, k],
                                    beta[0, j, k], sii, siu, siuu,
                                    ail, aii, aiu, bill, bil, bii, biu, biuu)
    elif axis == 1:
        for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                Biharmonic_matvec1D(v[i, :, k], b[i, :, k], a0, alfa[i, 0, k],
                                    beta[i, 0, k], sii, siu, siuu,
                                    ail, aii, aiu, bill, bil, bii, biu, biuu)
    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                Biharmonic_matvec1D(v[i, j], b[i, j], a0, alfa[i, j, 0],
                                    beta[i, j, 0], sii, siu, siuu,
                                    ail, aii, aiu, bill, bil, bii, biu, biuu)


@nb.jit(nopython=True, fastmath=True, cache=True)
def Biharmonic_matvec1D(v, b, a0, alfa, beta,
                        sii, siu, siuu,
                        ail, aii, aiu,
                        bill, bil, bii, biu, biuu):
    N = sii.shape[0]
    ldd = np.empty(N)
    ld = np.empty(N)
    dd = np.empty(N)
    ud = np.empty(N)
    udd = np.empty(N)

    for i in range(N):
        dd[i] = a0*sii[i] + alfa*aii[i] + beta*bii[i]

    for i in range(N-2):
        ld[i] = alfa*ail[i] + beta*bil[i]

    for i in range(N-4):
        ldd[i] = beta*bill[i]

    for i in range(N-2):
        ud[i] = a0*siu[i] + alfa*aiu[i] + beta*biu[i]

    for i in range(N-4):
        udd[i] = a0*siuu[i] + beta*biuu[i]

    i = N-1
    b[i] = ldd[i-4]*v[i-4]+ ld[i-2]* v[i-2] + dd[i]*v[i]
    i = N-2
    b[i] = ldd[i-4]*v[i-4]+ ld[i-2]* v[i-2] + dd[i]*v[i]
    i = N-3
    b[i] = ldd[i-4]*v[i-4]+ ld[i-2]* v[i-2] + dd[i]*v[i] + ud[i]*v[i+2]
    i = N-4
    b[i] = ldd[i-4]*v[i-4]+ ld[i-2]* v[i-2] + dd[i]*v[i] + ud[i]*v[i+2]
    i = N-5
    b[i] = ldd[i-4]*v[i-4]+ ld[i-2]* v[i-2] + dd[i]*v[i] + ud[i]*v[i+2] + udd[i]*v[i+4]
    i = N-6
    b[i] = ldd[i-4]*v[i-4]+ ld[i-2]* v[i-2] + dd[i]*v[i] + ud[i]*v[i+2] + udd[i]*v[i+4]

    s1 = 0.0
    s2 = 0.0
    o1 = 0.0
    o2 = 0.0
    for k in range(N-7, -1, -1):
        j = k+6
        p = k*sii[k]/(k+1.)
        r = 24*(k+1)*(k+2)*np.pi
        d = v[j]/(j+3.)
        if k % 2 == 0:
            s1 += d
            s2 += (j+2)*(j+2)*d
            b[k] = (p*s1 + r*s2)*a0
        else:
            o1 += d
            o2 += (j+2)*(j+2)*d
            b[k] = (p*o1 + r*o2)*a0

        if k > 3:
            b[k] += ldd[k-4]*v[k-4]+ ld[k-2]* v[k-2] + dd[k]*v[k] + ud[k]*v[k+2] + udd[k]*v[k+4]
        elif k > 1:
            b[k] += ld[k-2]* v[k-2] + dd[k]*v[k] + ud[k]*v[k+2] + udd[k]*v[k+4]
        else:
            b[k] += dd[k]*v[k] + ud[k]*v[k+2] + udd[k]*v[k+4]

def Biharmonic_matvec(v, b, a0, alfa, beta,
                      sii, siu, siuu, ail, aii, aiu,
                      bill, bil, bii, biu, biuu, axis):
    n = v.ndim
    if n == 1:
        Biharmonic_matvec1D(v, b, a0, alfa, beta,
                            sii, siu, siuu, ail, aii, aiu,
                            bill, bil, bii, biu, biuu)
    elif n == 2:
        Biharmonic_matvec2D(v, b, a0, alfa, beta,
                            sii, siu, siuu, ail, aii, aiu,
                            bill, bil, bii, biu, biuu, axis)
    elif n == 3:
        Biharmonic_matvec3D(v, b, a0, alfa, beta,
                            sii, siu, siuu, ail, aii, aiu,
                            bill, bil, bii, biu, biuu, axis)



r"""
This module contains linear algebra solvers for SparseMatrices,
TPMatrices and BlockMatrices.
"""
from numbers import Number, Integral
import numpy as np
from scipy.sparse import spmatrix, kron
from scipy.sparse.linalg import splu
from shenfun.config import config
from shenfun.optimization import optimizer, runtimeoptimizer
from shenfun.matrixbase import SparseMatrix, extract_bc_matrices, \
    BlockMatrix, get_simplified_tpmatrices
from shenfun.forms.arguments import Function
from mpi4py import MPI
comm = MPI.COMM_WORLD

def Solver(mats):
    """Return appropriate solver for `mats`

    Parameters
    ----------
    mats : :class:`.SparseMatrix` or list of :class:`.SparseMatrix` instances

    Returns
    -------
    Matrix solver (:class:`.SparseMatrixSolver`)

    Note
    ----
    The list of matrices may include boundary matrices. The returned solver
    will incorporate these boundary matrices automatically on the right hand
    side of the equation system.
    """
    assert isinstance(mats, (SparseMatrix, list))
    bc_mats = []
    mat = mats
    if isinstance(mats, list):
        bc_mats = extract_bc_matrices([mats])
        mat = sum(mats[1:], mats[0])
    return mat.get_solver()([mat]+bc_mats)

class SparseMatrixSolver:
    """Solver for :class:`.SparseMatrix` matrices.

    Parameters
    ----------
    mats : :class:`.SparseMatrix` or list of :class:`.SparseMatrix` instances

    Note
    ----
    The list of matrices may include boundary matrices. The returned solver
    will incorporate these boundary matrices automatically on the right hand
    side of the equation system.

    """
    def __init__(self, mats):
        assert isinstance(mats, (SparseMatrix, list))
        self.bc_mats = []
        if isinstance(mats, list):
            bc_mats = extract_bc_matrices([mats])
            mats = sum(mats[1:], mats[0])
            self.bc_mats = bc_mats
        self.mat = mats
        self._lu = None
        self._inner_arg = None # argument to inner_solve
        self.dtype = None
        assert self.mat.shape[0] == self.mat.shape[1]

    def apply_bcs(self, b, u, axis=0):
        if len(self.bc_mats) > 0:
            u.set_boundary_dofs()
            w0 = np.zeros_like(b)
            for bc_mat in self.bc_mats:
                b -= bc_mat.matvec(u, w0, axis=axis)
        return b

    def apply_constraints(self, b, constraints, axis=0):
        """Apply constraints to matrix `self.mat` and rhs vector `b`

        Parameters
        ----------
        b : array
        constraints : tuple of 2-tuples
            The 2-tuples represent (row, val)
            The constraint indents the matrix row and sets b[row] = val
        axis : int
            The axis we are solving over
        """
        # Only apply constraint to matrix first time around

        if len(constraints) > 0:
            if b.ndim > 1:
                T = b.function_space().bases[axis]

            A = self.mat
            if isinstance(A, spmatrix):
                for (row, val) in constraints:
                    if self._lu is None:
                        A = A.tolil()
                        _, zerorow = A[row].nonzero()
                        A[(row, zerorow)] = 0
                        A[row, row] = 1
                        self.mat = A.tocsc()

                    if b.ndim > 1:
                        b[T.si[row]] = val
                    else:
                        b[row] = val

            elif isinstance(A, SparseMatrix):
                for (row, val) in constraints:
                    if self._lu is None:
                        for key, vals in A.items():
                            if key >= 0:
                                M = A.shape[0]-key
                                v = np.broadcast_to(np.atleast_1d(vals), M).copy()
                                if row < M:
                                    v[row] = int(key == 0)/A.scale
                            elif key < 0:
                                M = A.shape[0]+key
                                v = np.broadcast_to(np.atleast_1d(vals), M).copy()
                                if row+key < M and row+key > 0:
                                    v[row+key] = 0
                            A[key] = v
                    if b.ndim > 1:
                        b[T.si[row]] = val
                    else:
                        b[row] = val
        return b

    def perform_lu(self):
        if self._lu is None:
            if isinstance(self.mat, SparseMatrix):
                self.mat = self.mat.diags('csc')
            self._lu = splu(self.mat, permc_spec=config['matrix']['sparse']['permc_spec'])
            self.dtype = self.mat.dtype.char
            self._inner_arg = (self._lu, self.dtype)
        return self._lu

    def solve(self, b, u, axis, lu):
        """Solve Au=b

        Solve along axis if b and u are multidimensional arrays.

        Parameters
        ----------
        b, u : arrays of rhs and output
            Both can be multidimensional
        axis : int
            The axis we are solving over
        lu : LU-decomposition
            Can be either the output from splu, or a dia-matrix containing
            the L and U matrices. The latter is used in subclasses.

        """
        if axis > 0:
            u = np.moveaxis(u, axis, 0)
            if u is not b:
                b = np.moveaxis(b, axis, 0)

        s = slice(0, self.mat.shape[0])
        if b.ndim == 1:
            if b.dtype.char in 'fdg' or self.dtype in 'FDG':
                u[s] = lu.solve(b[s])

            else:
                u.real[s] = lu.solve(b[s].real)
                u.imag[s] = lu.solve(b[s].imag)

        else:
            N = b[s].shape[0]
            P = np.prod(b[s].shape[1:])
            br = b[s].reshape((N, P))
            if b.dtype.char in 'fdg' or self.dtype in 'FDG':
                u[s] = lu.solve(br).reshape(u[s].shape)
            else:
                u.real[s] = lu.solve(br.real).reshape(u[s].shape)
                u.imag[s] = lu.solve(br.imag).reshape(u[s].shape)

        if axis > 0:
            u = np.moveaxis(u, 0, axis)
            if u is not b:
                b = np.moveaxis(b, 0, axis)

        return u

    @staticmethod
    def inner_solve(u, lu):
        """Solve Au=b for one-dimensional u

        On entry u is the rhs b, on exit it contains the solution.

        Parameters
        ----------
        u : array 1D
            rhs on entry and solution on exit
        lu : LU-decomposition
            Can be either a 2-tuple with (output from splu, dtype), or a scipy
            dia-matrix containing the L and U matrices. The latter is used in
            subclasses.

        """
        lu, dtype = lu
        s = slice(0, lu.shape[0])
        if u.dtype.char in 'fdg' or dtype in 'FDG':
            u[s] = lu.solve(u[s])

        else:
            u.real[s] = lu.solve(u.real[s])
            u.imag[s] = lu.solve(u.imag[s])

    def __call__(self, b, u=None, axis=0, constraints=()):
        """Solve matrix problem Au = b along axis

        This routine also applies boundary conditions and constraints,
        and performes LU-decomposition on the fully assembled matrix.

        Parameters
        ----------
        b : array
            Array of right hand side on entry and solution on exit unless
            u is provided.
        u : array, optional
            Output array
        axis : int, optional
            The axis over which to solve for if b and u are multidimensional
        constraints : tuple of 2-tuples
            The 2-tuples represent (row, val)
            The constraint indents the matrix row and sets b[row] = val

        Note
        ----
        If u is not provided, then b is overwritten with the solution and returned

        """
        if u is None:
            u = b
        b = self.apply_bcs(b, u, axis=axis)
        b = self.apply_constraints(b, constraints, axis=axis)
        lu = self.perform_lu() # LU must be performed after constraints, because constraints modify the matrix
        u = self.solve(b, u, axis, lu)
        if hasattr(u, 'set_boundary_dofs'):
            u.set_boundary_dofs()
        return u

class BandedMatrixSolver(SparseMatrixSolver):
    def __init__(self, mats):
        SparseMatrixSolver.__init__(self, mats)
        self._lu = self.mat.diags('dia')

    def solve(self, b, u, axis, lu):
        if u is not b:
            sl = u.function_space().slice() if hasattr(u, 'function_space') else slice(None)
            u[sl] = b[sl]
        self.Solve(u, lu.data, axis)
        return u

    @staticmethod
    def LU(data):
        """LU-decomposition using either Cython or Numba

        Parameters
        ----------
        data : 2D-array
            Storage for dia-matrix on entry and L and U matrices
            on exit.
        """
        raise NotImplementedError

    @staticmethod
    def Solve(u, data, axis):
        """Fast solve using either Cython or Numba

        Parameters
        ----------
        u : array
            rhs on entry, solution on exit
        data : 2D-array
            Storage for dia-matrix containing L and U matrices
        axis : int
            The axis we are solving over
        """
        raise NotImplementedError


class DiagMA(BandedMatrixSolver):
    """Diagonal matrix solver

    Parameters
    ----------
    mats : Diagonal :class:`.SparseMatrix` or list of diagonal :class:`.SparseMatrix` instances

    """
    def __init__(self, mats):
        BandedMatrixSolver.__init__(self, mats)
        self.issymmetric = True
        self._inner_arg = self._lu.data

    def perform_lu(self):
        return self._lu

    def apply_constraints(self, b, constraints, axis=0):
        if len(constraints) > 0:
            assert len(constraints) == 1
            assert constraints[0][0] == 0, 'Can only fix first row'
            self._lu.diagonal(0)[0] = 1
            s = [slice(None)]*len(b.shape)
            s[axis] = 0
            b[tuple(s)] = constraints[0][1]
        return b

    @staticmethod
    def inner_solve(u, lu):
        d = lu[0]
        u[:d.shape[0]] /= d

    @staticmethod
    @runtimeoptimizer
    def Solve(u, data, axis):
        if axis > 0:
            u = np.moveaxis(u, axis, 0)
        DiagMA.inner_solve(u, data)
        if axis > 0:
            u = np.moveaxis(u, 0, axis)

class TDMA(BandedMatrixSolver):
    """Tridiagonal matrix solver

    Parameters
    ----------
    mats : :class:`.SparseMatrix` or list of :class:`.SparseMatrix` instances
        Tridiagonal matrix with diagonals in offsets -2, 0, 2

    """
    def __init__(self, mats):
        BandedMatrixSolver.__init__(self, mats)
        self.issymmetric = self.mat.issymmetric

    @staticmethod
    @runtimeoptimizer
    def LU(data):
        ld = data[0, :-2]
        d = data[1, :]
        ud = data[2, 2:]
        n = d.shape[0]
        for i in range(2, n):
            ld[i-2] = ld[i-2]/d[i-2]
            d[i] = d[i] - ld[i-2]*ud[i-2]

    def apply_constraints(self, b, constraints, axis=0):
        if len(constraints) > 0:
            assert len(constraints) == 1
            assert constraints[0][0] == 0, 'Can only fix first row of TDMA'
            self._lu.diagonal(0)[0] = 1
            self._lu.diagonal(2)[0] = 0
            s = [slice(None)]*len(b.shape)
            s[axis] = 0
            b[tuple(s)] = constraints[0][1]
        return b

    def perform_lu(self):
        if self._inner_arg is None:
            self.LU(self._lu.data)
            self._inner_arg = self._lu.data
        return self._lu

    @staticmethod
    def inner_solve(u, data):
        ld = data[0, :-2]
        d = data[1, :]
        ud = data[2, 2:]
        n = d.shape[0]
        for i in range(2, n):
            u[i] -= ld[i-2]*u[i-2]

        u[n-1] = u[n-1]/d[n-1]
        u[n-2] = u[n-2]/d[n-2]
        for i in range(n - 3, -1, -1):
            u[i] = (u[i] - ud[i]*u[i+2])/d[i]

    @staticmethod
    @runtimeoptimizer
    def Solve(u, data, axis):
        if axis > 0:
            u = np.moveaxis(u, axis, 0)
        TDMA.inner_solve(u, data)
        if axis > 0:
            u = np.moveaxis(u, 0, axis)


class TDMA_O(BandedMatrixSolver):
    """Tridiagonal matrix solver

    Parameters
    ----------
    mats : :class:`.SparseMatrix` or list of :class:`.SparseMatrix` instances
        Symmetric tridiagonal matrix with diagonals in offsets -1, 0, 1

    """
    # pylint: disable=too-few-public-methods

    def __init__(self, mats):
        BandedMatrixSolver.__init__(self, mats)

    def perform_lu(self):
        if self._inner_arg is None:
            self.LU(self._lu.data)
            self._inner_arg = self._lu.data
        return self._lu

    @staticmethod
    @runtimeoptimizer
    def LU(data):
        ld = data[0, :-1]
        d = data[1, :]
        ud = data[2, 1:]
        n = d.shape[0]
        for i in range(1, n):
            ld[i-1] = ld[i-1]/d[i-1]
            d[i] -= ld[i-1]*ud[i-1]

    @staticmethod
    def inner_solve(u, data):
        ld = data[0, :-1]
        d = data[1, :]
        ud = data[2, 1:]
        n = d.shape[0]
        for i in range(1, n):
            u[i] -= ld[i-1]*u[i-1]

        u[n-1] = u[n-1]/d[n-1]
        for i in range(n-2, -1, -1):
            u[i] = (u[i] - ud[i]*u[i+1])/d[i]

    @staticmethod
    @runtimeoptimizer
    def Solve(u, data, axis):
        if axis > 0:
            u = np.moveaxis(u, axis, 0)
        TDMA_O.inner_solve(u, data)
        if axis > 0:
            u = np.moveaxis(u, 0, axis)


class PDMA(BandedMatrixSolver):
    """Pentadiagonal matrix solver

    Parameters
    ----------
    mats : :class:`.SparseMatrix` or list of :class:`.SparseMatrix` instances
        Pentadiagonal matrix with diagonals in offsets
        -4, -2, 0, 2, 4

    """
    def __init__(self, mats):
        BandedMatrixSolver.__init__(self, mats)
        assert len(self.mat) == 5

    def apply_constraints(self, b, constraints, axis=0):
        if len(constraints) > 0:
            assert len(constraints) == 1
            assert constraints[0][0] == 0, 'Can only fix first row of PDMA'
            self._lu.diagonal(0)[0] = 1
            self._lu.diagonal(2)[0] = 0
            self._lu.diagonal(4)[0] = 0
            if b.ndim > 1:
                s = [slice(None)]*len(b.shape)
                s[axis] = 0
                b[tuple(s)] = constraints[0][1]
            else:
                b[0] = constraints[0][1]
            self._inner_arg = self._lu.data
        return b

    @staticmethod
    @runtimeoptimizer
    def LU(data): # pragma: no cover
        a = data[0, :-4]
        b = data[1, :-2]
        d = data[2, :]
        e = data[3, 2:]
        f = data[4, 4:]
        n = d.shape[0]
        m = e.shape[0]
        k = n - m

        for i in range(n-2*k):
            lam = b[i]/d[i]
            d[i+k] -= lam*e[i]
            e[i+k] -= lam*f[i]
            b[i] = lam
            lam = a[i]/d[i]
            b[i+k] -= lam*e[i]
            d[i+2*k] -= lam*f[i]
            a[i] = lam

        i = n-4
        lam = b[i]/d[i]
        d[i+k] -= lam*e[i]
        b[i] = lam
        i = n-3
        lam = b[i]/d[i]
        d[i+k] -= lam*e[i]
        b[i] = lam

    def perform_lu(self):
        if self._inner_arg is None:
            self.LU(self._lu.data)
            self._inner_arg = self._lu.data
        return self._lu

    @staticmethod
    def inner_solve(u, data):
        a = data[0, :-4]
        b = data[1, :-2]
        d = data[2, :]
        e = data[3, 2:]
        f = data[4, 4:]
        n = d.shape[0]
        u[2] -= b[0]*u[0]
        u[3] -= b[1]*u[1]
        for k in range(4, n):
            u[k] -= (b[k-2]*u[k-2] + a[k-4]*u[k-4])
        u[n-1] /= d[n-1]
        u[n-2] /= d[n-2]
        u[n-3] = (u[n-3]-e[n-3]*u[n-1])/d[n-3]
        u[n-4] = (u[n-4]-e[n-4]*u[n-2])/d[n-4]
        for k in range(n-5, -1, -1):
            u[k] = (u[k]-e[k]*u[k+2]-f[k]*u[k+4])/d[k]

    @staticmethod
    @runtimeoptimizer
    def Solve(u, data, axis):
        if axis > 0:
            u = np.moveaxis(u, axis, 0)
        PDMA.inner_solve(u, data)
        if axis > 0:
            u = np.moveaxis(u, 0, axis)

class FDMA(BandedMatrixSolver):
    """4-diagonal matrix solver

    Parameters
    ----------
    mats : :class:`.SparseMatrix` or list of :class:`.SparseMatrix` instances
        4-diagonal matrix with diagonals in offsets -2, 0, 2, 4

    """
    # pylint: disable=too-few-public-methods
    def __init__(self, mats):
        BandedMatrixSolver.__init__(self, mats)

    def perform_lu(self):
        if self._inner_arg is None:
            self.LU(self._lu.data)
            self._inner_arg = self._lu.data
        return self._lu

    @staticmethod
    @runtimeoptimizer
    def LU(data):
        ld = data[0, :-2]
        d = data[1, :]
        u1 = data[2, 2:]
        u2 = data[3, 4:]
        n = d.shape[0]
        for i in range(2, n):
            ld[i-2] = ld[i-2]/d[i-2]
            d[i] = d[i] - ld[i-2]*u1[i-2]
            if i < n-2:
                u1[i] = u1[i] - ld[i-2]*u2[i-2]

    def apply_constraints(self, b, constraints, axis=0):
        if len(constraints) > 0:
            assert len(constraints) == 1
            assert constraints[0][0] == 0, 'Can only fix first row of FDMA'
            self._lu.diagonal(0)[0] = 1
            self._lu.diagonal(2)[0] = 0
            self._lu.diagonal(4)[0] = 0
            s = [slice(None)]*len(b.shape)
            s[axis] = 0
            b[tuple(s)] = constraints[0][1]
        return b

    @staticmethod
    def inner_solve(u, data):
        ld = data[0, :-2]
        d = data[1, :]
        u1 = data[2, 2:]
        u2 = data[3, 4:]
        n = d.shape[0]
        for i in range(2, n):
            u[i] -= ld[i-2]*u[i-2]
        u[n-1] = u[n-1]/d[n-1]
        u[n-2] = u[n-2]/d[n-2]
        u[n-3] = (u[n-3] - u1[n-3]*u[n-1])/d[n-3]
        u[n-4] = (u[n-4] - u1[n-4]*u[n-2])/d[n-4]
        for i in range(n - 5, -1, -1):
            u[i] = (u[i] - u1[i]*u[i+2] - u2[i]*u[i+4])/d[i]

    @staticmethod
    @runtimeoptimizer
    def Solve(u, data, axis):
        if axis > 0:
            u = np.moveaxis(u, axis, 0)
        FDMA.inner_solve(u, data)
        if axis > 0:
            u = np.moveaxis(u, 0, axis)

class TwoDMA(BandedMatrixSolver):
    """2-diagonal matrix solver

    Parameters
    ----------
    mats : :class:`.SparseMatrix` or list of :class:`.SparseMatrix` instances
        2-diagonal matrix with diagonals in offsets 0, 2

    """
    def __init__(self, mats):
        BandedMatrixSolver.__init__(self, mats)
        self._inner_arg = self._lu.data

    def apply_constraints(self, b, constraints, axis=0):
        if len(constraints) > 0:
            assert len(constraints) == 1
            assert constraints[0][0] == 0, 'Can only fix first row of TwoDMA'
            self._lu.diagonal(0)[0] = 1
            self._lu.diagonal(2)[0] = 0
            s = [slice(None)]*len(b.shape)
            s[axis] = 0
            b[tuple(s)] = constraints[0][1]
        return b

    def perform_lu(self):
        return self._lu

    @staticmethod
    def inner_solve(u, data):
        d = data[0, :]
        u1 = data[1, 2:]
        n = d.shape[0]
        u[n-1] = u[n-1]/d[n-1]
        u[n-2] = u[n-2]/d[n-2]
        for i in range(n - 3, -1, -1):
            u[i] = (u[i] - u1[i]*u[i+2])/d[i]

    @staticmethod
    @runtimeoptimizer
    def Solve(u, data, axis):
        if axis > 0:
            u = np.moveaxis(u, axis, 0)
        TwoDMA.inner_solve(u, data)
        if axis > 0:
            u = np.moveaxis(u, 0, axis)

class ThreeDMA(BandedMatrixSolver):
    """3-diagonal matrix solver - all diagonals upper

    Parameters
    ----------
    mats : :class:`.SparseMatrix` or list of :class:`.SparseMatrix` instances
        3-diagonal matrix with diagonals in offsets 0, 2, 4

    """
    def __init__(self, mats):
        BandedMatrixSolver.__init__(self, mats)
        self._inner_arg = self._lu.data

    def apply_constraints(self, b, constraints, axis=0):
        if len(constraints) > 0:
            assert len(constraints) == 1
            assert constraints[0][0] == 0, 'Can only fix first row of TwoDMA'
            self._lu.diagonal(0)[0] = 1
            self._lu.diagonal(2)[0] = 0
            self._lu.diagonal(4)[0] = 0
            s = [slice(None)]*len(b.shape)
            s[axis] = 0
            b[tuple(s)] = constraints[0][1]
        return b

    def perform_lu(self):
        return self._lu

    @staticmethod
    def inner_solve(u, data):
        d = data[0, :]
        u1 = data[1, 2:]
        u2 = data[1, 4:]
        n = d.shape[0]
        u[n-1] = u[n-1]/d[n-1]
        u[n-2] = u[n-2]/d[n-2]
        u[n-3] = (u[n-3]-u1[n-3]*u[n-1])/d[n-3]
        u[n-4] = (u[n-4]-u1[n-4]*u[n-2])/d[n-4]
        for i in range(n - 5, -1, -1):
            u[i] = (u[i] - u1[i]*u[i+2] - u2[i]*u[i+4])/d[i]

    @staticmethod
    @runtimeoptimizer
    def Solve(u, data, axis):
        if axis > 0:
            u = np.moveaxis(u, axis, 0)
        ThreeDMA.inner_solve(u, data)
        if axis > 0:
            u = np.moveaxis(u, 0, axis)

class HeptaDMA(BandedMatrixSolver):
    """Heptadiagonal matrix solver

    Parameters
    ----------
    mats : :class:`.SparseMatrix` or list of :class:`.SparseMatrix` instances
        Heptadiagonal matrix with diagonals in offsets
        -4, -2, 0, 2, 4, 6, 8

    """
    def __init__(self, mats):
        BandedMatrixSolver.__init__(self, mats)
        assert len(self.mat) == 7

    def apply_constraints(self, b, constraints, axis=0):
        if len(constraints) > 0:
            assert len(constraints) == 1
            assert constraints[0][0] == 0, 'Can only fix first row of HeptaDMA'
            self._lu.diagonal(0)[0] = 1
            self._lu.diagonal(2)[0] = 0
            self._lu.diagonal(4)[0] = 0
            self._lu.diagonal(6)[0] = 0
            self._lu.diagonal(8)[0] = 0
            if b.ndim > 1:
                s = [slice(None)]*len(b.shape)
                s[axis] = 0
                b[tuple(s)] = constraints[0][1]
            else:
                b[0] = constraints[0][1]
            self._inner_arg = self._lu.data
        return b

    @staticmethod
    @runtimeoptimizer
    def LU(data): # pragma: no cover
        a = data[0, :-4]
        b = data[1, :-2]
        d = data[2, :]
        e = data[3, 2:]
        f = data[4, 4:]
        g = data[5, 6:]
        h = data[6, 8:]
        n = d.shape[0]
        m = e.shape[0]
        k = n - m
        for i in range(n-2*k):
            lam = b[i]/d[i]
            d[i+k] -= lam*e[i]
            e[i+k] -= lam*f[i]
            if i < n-6:
                f[i+k] -= lam*g[i]
            if i < n-8:
                g[i+k] -= lam*h[i]
            b[i] = lam
            lam = a[i]/d[i]
            b[i+k] -= lam*e[i]
            d[i+2*k] -= lam*f[i]
            if i < n-6:
                e[i+2*k] -= lam*g[i]
            if i < n-8:
                f[i+2*k] -= lam*h[i]
            a[i] = lam
        i = n-4
        lam = b[i]/d[i]
        d[i+k] -= lam*e[i]
        b[i] = lam
        i = n-3
        lam = b[i]/d[i]
        d[i+k] -= lam*e[i]
        b[i] = lam

    def perform_lu(self):
        if self._inner_arg is None:
            self.LU(self._lu.data)
            self._inner_arg = self._lu.data
        return self._lu

    @staticmethod
    def inner_solve(u, data):
        a = data[0, :-4]
        b = data[1, :-2]
        d = data[2, :]
        e = data[3, 2:]
        f = data[4, 4:]
        g = data[5, 6:]
        h = data[6, 8:]
        n = d.shape[0]
        u[2] -= b[0]*u[0]
        u[3] -= b[1]*u[1]
        for k in range(4, n):
            u[k] -= (b[k-2]*u[k-2] + a[k-4]*u[k-4])
        u[n-1] /= d[n-1]
        u[n-2] /= d[n-2]
        u[n-3] = (u[n-3]-e[n-3]*u[n-1])/d[n-3]
        u[n-4] = (u[n-4]-e[n-4]*u[n-2])/d[n-4]
        u[n-5] = (u[n-5]-e[n-5]*u[n-3]-f[n-5]*u[n-1])/d[n-5]
        u[n-6] = (u[n-6]-e[n-6]*u[n-4]-f[n-6]*u[n-2])/d[n-6]
        u[n-7] = (u[n-7]-e[n-7]*u[n-5]-f[n-7]*u[n-3]-g[n-7]*u[n-1])/d[n-7]
        u[n-8] = (u[n-8]-e[n-8]*u[n-6]-f[n-8]*u[n-4]-g[n-8]*u[n-2])/d[n-8]
        for k in range(n-9, -1, -1):
            u[k] = (u[k]-e[k]*u[k+2]-f[k]*u[k+4]-g[k]*u[k+6]-h[k]*u[k+8])/d[k]

    @staticmethod
    @runtimeoptimizer
    def Solve(u, data, axis):
        if axis > 0:
            u = np.moveaxis(u, axis, 0)
        HeptaDMA.inner_solve(u, data)
        if axis > 0:
            u = np.moveaxis(u, 0, axis)


class Solve(SparseMatrixSolver):
    """Generic solver class for ::class:`.SparseMatrix`

    Possibly with inhomogeneous boundary values

    Parameters
    ----------
    mats : :class:`.SparseMatrix` or list of :class:`.SparseMatrix` instances
    format : str, optional
        The format of the scipy.sparse.spmatrix to convert into
        before solving. Default is Compressed Sparse Column `csc`.

    Note
    ----
    This solver converts the matrix to a Scipy sparse matrix of choice and
    uses `scipy.sparse` methods `splu` and `spsolve`.

    """
    def __init__(self, mats, format=None):
        format = config['matrix']['sparse']['solve'] if format is None else format
        SparseMatrixSolver.__init__(self, mats)
        self.mat = self.mat.diags(format)


class SolverGeneric2ND:
    """Generic solver for problems consisting of tensorproduct matrices
    containing two non-diagonal submatrices.

    Parameters
    ----------
    mats : sequence
        sequence of instances of :class:`.TPMatrix`

    Note
    ----
    In addition to two non-diagonal matrices, the solver can also handle one
    additional diagonal matrix (one Fourier matrix).
    """

    def __init__(self, tpmats):
        tpmats = get_simplified_tpmatrices(tpmats)
        bc_mats = extract_bc_matrices([tpmats])
        self.tpmats = tpmats
        self.bc_mats = bc_mats
        self.T = tpmats[0].space
        self.mats2D = {}
        self._lu = None

    def matvec(self, u, c):
        c.fill(0)
        if u.ndim == 2:
            s0 = tuple(base.slice() for base in self.T)
            c[s0] = self.mats2D.dot(u[s0].flatten()).reshape(self.T.dims())
        else:
            raise NotImplementedError
        return c

    def get_diagonal_axis(self):
        naxes = self.T.get_nondiagonal_axes()
        diagonal_axis = np.setxor1d([0, 1, 2], naxes)
        assert len(diagonal_axis) == 1
        return diagonal_axis[0]

    def diags(self, i):
        """Return matrix for given index `i` in diagonal direction

        Parameters
        ----------
        i : int
            Fourier wavenumber
        """
        if i in self.mats2D:
            return self.mats2D[i]

        if self.T.dimensions == 2:
            # In 2D there's just 1 matrix, store and reuse
            m = self.tpmats[0]
            M0 = m.diags('csc')
            for m in self.tpmats[1:]:
                M0 = M0 + m.diags('csc')

        else:
            # 1 matrix per Fourier coefficient
            naxes = self.T.get_nondiagonal_axes()
            m = self.tpmats[0]
            diagonal_axis = self.get_diagonal_axis()
            sc = [0, 0, 0]
            sc[diagonal_axis] = i if m.scale.shape[diagonal_axis] > 1 else 0
            A0 = m.mats[naxes[0]].diags('csc')
            A1 = m.mats[naxes[1]].diags('csc')
            M0 = kron(A0, A1, 'csc')
            M0 *= m.scale[tuple(sc)]
            for m in self.tpmats[1:]:
                A0 = m.mats[naxes[0]].diags('csc')
                A1 = m.mats[naxes[1]].diags('csc')
                M1 = kron(A0, A1, 'csc')
                sc[diagonal_axis] = i if m.scale.shape[diagonal_axis] > 1 else 0
                M1 *= m.scale[tuple(sc)]
                M0 = M0 + M1
        self.mats2D[i] = M0
        return M0

    def apply_constraints(self, b, constraints):
        if len(constraints) > 0:
            if self._lu is None:
                A = self.mats2D[0]
                A = A.tolil()
                for (row, val) in constraints:
                    _, zerorow = A[row].nonzero()
                    A[(row, zerorow)] = 0
                    A[row, row] = 1
                    b[row] = val
                self.mats2D[0] = A.tocsc()
            else:
                for (row, val) in constraints:
                    b[row] = val
        return b

    def assemble(self):
        if len(self.mats2D) == 0:
            ndim = self.tpmats[0].dimensions
            if ndim == 2:
                mat = self.diags(0)
                self.mats2D[0] = mat

            elif ndim == 3:
                diagonal_axis = self.get_diagonal_axis()
                for i in range(self.T.shape(True)[diagonal_axis]):
                    M0 = self.diags(i)
                    self.mats2D[i] = M0
        return self.mats2D

    def perform_lu(self):
        if self._lu is not None:
            return self._lu
        ndim = self.tpmats[0].dimensions
        self._lu = {}
        if ndim == 2:
            self._lu[0] = splu(self.mats2D[0], permc_spec=config['matrix']['sparse']['permc_spec'])
        else:
            diagonal_axis = self.get_diagonal_axis()
            for i in range(self.T.shape(True)[diagonal_axis]):
                self._lu[i] = splu(self.mats2D[i], permc_spec=config['matrix']['sparse']['permc_spec'])
        return self._lu

    def __call__(self, b, u=None, constraints=()):
        """Solve generic problem

        Parameters
        ----------
        b : array, right hand side
        u : array, solution
        constraints : tuple of 2-tuples
            Each 2-tuple (row, value) is a constraint set for the non-periodic
            direction, for Fourier index 0 in 2D and (0, 0) in 3D

        """
        if u is None:
            u = b
        else:
            assert u.shape == b.shape

        if len(self.bc_mats) > 0:
            u.set_boundary_dofs()
            w0 = Function(self.T).v
            for bc_mat in self.bc_mats:
                b -= bc_mat.matvec(u, w0)

        mats = self.assemble()

        b = self.apply_constraints(b, constraints)

        lu = self.perform_lu()

        if u.ndim == 2:
            s0 = self.T.slice()
            bs = b[s0].flatten()
            if b.dtype.char in 'fdg' or self.mats2D[0].dtype.char in 'FDG':
                u[s0] = lu[0].solve(bs).reshape(self.T.dims())
            else:
                u.real[s0] = lu[0].solve(bs.real).reshape(self.T.dims())
                u.imag[s0] = lu[0].solve(bs.imag).reshape(self.T.dims())

        elif u.ndim == 3:
            naxes = self.T.get_nondiagonal_axes()
            diagonal_axis = self.get_diagonal_axis()
            s0 = list(self.T.slice())
            for i in range(self.T.shape(True)[diagonal_axis]):
                s0[diagonal_axis] = i
                bs = b[tuple(s0)].flatten()
                shape = np.take(self.T.dims(), naxes)
                if b.dtype.char in 'fdg' or self.mats2D[0].dtype.char in 'FDG':
                    u[tuple(s0)] = lu[i].solve(bs).reshape(shape)
                else:
                    u.real[tuple(s0)] = lu[i].solve(bs.real).reshape(shape)
                    u.imag[tuple(s0)] = lu[i].solve(bs.imag).reshape(shape)

        if hasattr(u, 'set_boundary_dofs'):
            u.set_boundary_dofs()
        return u

class SolverDiagonal:
    """Solver for purely diagonal matrices, like Fourier in Cartesian coordinates.

    Parameters
    ----------
    tpmats : sequence
        sequence of instances of :class:`.TPMatrix`
    """
    def __init__(self, tpmats):
        tpmats = get_simplified_tpmatrices(tpmats)
        assert len(tpmats) == 1
        self.mat = tpmats[0]

    def __call__(self, b, u=None, constraints=()):
        """Solve problem with :class:`.TPMatrix` consisting only of diagonal
        matrices

        Parameters
        ----------
        b : array, right hand side
        u : array, solution
        constraints : tuple of 2-tuples
            Each 2-tuple (row, value) is a constraint set for the non-periodic
            direction, for Fourier index 0 in 2D and (0, 0) in 3D

        """
        return self.mat.solve(b, u=u, constraints=constraints)

class Solver2D:
    """Generic solver for tensorproductspaces in 2D

    Parameters
    ----------
    mats : sequence
        sequence of instances of :class:`.TPMatrix`

    Note
    ----
    If there are boundary matrices in the list of mats, then
    these matrices are used to modify the right hand side before
    solving. If this is not the desired behaviour, then use
    :func:`.extract_bc_matrices` on tpmats before using this class.

    """

    def __init__(self, tpmats):
        bc_mats = extract_bc_matrices([tpmats])
        self.tpmats = tpmats
        self.bc_mats = bc_mats
        self._lu = None
        m = tpmats[0]
        self.T = m.space
        assert m._issimplified is False, "Cannot use simplified matrices with this solver"
        mat = m.diags(format='csc')
        for m in tpmats[1:]:
            mat = mat + m.diags('csc')
        self.mat = mat

    def matvec(self, u, c):
        c.fill(0)
        s0 = tuple(base.slice() for base in self.T)
        c[s0] = self.mat.dot(u[s0].flatten()).reshape(self.T.dims())
        return c

    @staticmethod
    def apply_constraints(A, b, constraints):
        if len(constraints) > 0:
            A = A.tolil()
            for (row, val) in constraints:
                _, zerorow = A[row].nonzero()
                A[(row, zerorow)] = 0
                A[row, row] = 1
                b[row] = val
            A = A.tocsc()
        return A, b

    def __call__(self, b, u=None, constraints=()):
        """Solve generic problem for sum of :class:`TPMatrix` instances

        Parameters
        ----------
        b : array, right hand side
        u : array, solution
        constraints : tuple of 2-tuples
            Each 2-tuple (row, value) is a constraint set for the non-periodic
            direction, for Fourier index 0 in 2D and (0, 0) in 3D

        """
        if u is None:
            u = b
        else:
            pass
            #assert u.shape == b.shape

        if len(self.bc_mats) > 0:
            u.set_boundary_dofs()
            w0 = Function(self.T).v
            for bc_mat in self.bc_mats:
                b -= bc_mat.matvec(u, w0)

        s0 = tuple(base.slice() for base in self.T)
        assert b.dtype.char == u.dtype.char

        bs = b[s0].flatten()
        self.mat, bs = self.apply_constraints(self.mat, bs, constraints)
        if self._lu is None:
            self._lu = splu(self.mat, permc_spec=config['matrix']['sparse']['permc_spec'])

        if b.dtype.char in 'fdg' or self.mat.dtype.char in 'FDG':
            u[s0] = self._lu.solve(bs).reshape(self.T.dims())

        else:
            u.imag[s0] = self._lu.solve(bs.imag).reshape(self.T.dims())
            u.real[s0] = self._lu.solve(bs.real).reshape(self.T.dims())

        if hasattr(u, 'set_boundary_dofs'):
            u.set_boundary_dofs()
        return u

class Solver3D(Solver2D):
    """Generic solver for tensorproductspaces in 3D

    Parameters
    ----------
    mats : sequence
        sequence of instances of :class:`.TPMatrix`

    Note
    ----
    If there are boundary matrices in the list of mats, then
    these matrices are used to modify the right hand side before
    solving. If this is not the desired behaviour, then use
    :func:`.extract_bc_matrices` on mats before using this class.

    """

    def __init__(self, tpmats):
        Solver2D.__init__(self, tpmats)

class SolverND(Solver2D):
    """Generic solver for tensorproductspaces in N dimensions

    Parameters
    ----------
    mats : sequence
        sequence of instances of :class:`.TPMatrix`

    Note
    ----
    If there are boundary matrices in the list of mats, then
    these matrices are used to modify the right hand side before
    solving. If this is not the desired behaviour, then use
    :func:`.extract_bc_matrices` on mats before using this class.

    """

    def __init__(self, tpmats):
        Solver2D.__init__(self, tpmats)


class SolverGeneric1ND:
    """Generic solver for tensorproduct matrices consisting of
    non-diagonal matrices along only one axis and Fourier along
    the others.

    Parameters
    ----------
    mats : sequence
        sequence of instances of :class:`.TPMatrix`

    Note
    ----
    In addition to the one non-diagonal direction, the solver can also handle
    up to two diagonal (Fourier) directions.
    Also note that if there are boundary matrices in the list of mats, then
    these matrices are used to modify the right hand side before
    solving. If this is not the desired behaviour, then use
    :func:`.extract_bc_matrices` on mats before using this class.

    """

    def __init__(self, mats):
        assert isinstance(mats, list)
        mats = get_simplified_tpmatrices(mats)
        assert len(mats[0].naxes) == 1
        self.naxes = mats[0].naxes[0]
        bc_mats = extract_bc_matrices([mats])
        self.mats = mats
        self.testspace = mats[0].space
        self.trialspace = mats[0].trialspace
        self.bc_mats = bc_mats
        self.solvers1D = None
        self.assemble()
        self._lu = False
        self._data = None

    def matvec(self, u, c):
        c.fill(0)
        w0 = np.zeros_like(c)
        for mat in self.mats:
            c += mat.matvec(u, w0)

        if len(self.bc_mats) > 0:
            u.set_boundary_dofs()
            for bc_mat in self.bc_mats:
                c += bc_mat.matvec(u, w0)
        return c

    def assemble(self):
        ndim = self.mats[0].dimensions
        shape = self.mats[0].space.shape(True)
        self.solvers1D = []
        if ndim == 2:
            zi = np.ndindex((1, shape[1])) if self.naxes == 0 else np.ndindex((shape[0], 1))
            other_axis = (self.naxes+1) % 2
            for i in zi:
                sol = None
                for mat in self.mats:
                    sc = mat.scale[i] if mat.scale.shape[other_axis] > 1 else mat.scale[0, 0]
                    if sol:
                        sol += mat.mats[self.naxes]*sc
                    else:
                        sol = mat.mats[self.naxes]*sc
                self.solvers1D.append(Solver(sol))

        elif ndim == 3:
            s = [0, 0, 0]
            n0, n1 = np.setxor1d((0, 1, 2), self.naxes)
            for i in range(shape[n0]):
                self.solvers1D.append([])
                s[n0] = i
                for j in range(shape[n1]):
                    sol = None
                    s[n1] = j
                    for mat in self.mats:
                        sc = np.broadcast_to(mat.scale, shape)[tuple(s)]
                        if sol:
                            sol += mat.mats[self.naxes]*sc
                        else:
                            sol = mat.mats[self.naxes]*sc
                    self.solvers1D[-1].append(Solver(sol))

    def apply_constraints(self, b, constraints=()):
        #The SolverGeneric1ND solver can only constrain the first dofs of
        #the diagonal axes. For Fourier this is the zero dof with the
        #constant basis function exp(0).
        if constraints == ():
            return b
        ndim = self.mats[0].dimensions
        space = self.mats[0].space
        z0 = space.local_slice()
        paxes = np.setxor1d(range(ndim), self.naxes)
        s = [0]*ndim
        s[self.naxes] = slice(None)
        s = tuple(s)
        is_rank_zero = np.array([z0[i].start for i in paxes]).prod()
        sol = self.solvers1D[0] if ndim == 2 else self.solvers1D[0][0]
        if is_rank_zero != 0:
            return b
        sol.apply_constraints(b[s], constraints)
        return b

    def perform_lu(self):
        if self._lu is True:
            return

        if isinstance(self.solvers1D[0], SparseMatrixSolver):
            for m in self.solvers1D:
                lu = m.perform_lu()

        else:
            for mi in self.solvers1D:
                for mij in mi:
                    lu = mij.perform_lu()
        self._lu = True

    def get_data(self, is_rank_zero):
        if not self._data is None:
            return self._data

        if self.mats[0].dimensions == 2:
            data = np.zeros((len(self.solvers1D),)+self.solvers1D[-1]._inner_arg.shape)
            for i, sol in enumerate(self.solvers1D):
                if i == 0 and is_rank_zero:
                    continue
                else:
                    data[i] = sol._inner_arg
        elif self.mats[0].dimensions == 3:
            data = np.zeros((len(self.solvers1D), len(self.solvers1D[0]))+self.solvers1D[-1][-1]._inner_arg.shape)
            for i, m in enumerate(self.solvers1D):
                for j, sol in enumerate(m):
                    if i == 0 and j == 0 and is_rank_zero:
                        continue
                    else:
                        data[i, j] = sol._inner_arg
        self._data = data
        return data

    @staticmethod
    @runtimeoptimizer
    def solve_data(u, data, sol, naxes, is_rank_zero):
        s = [0]*u.ndim
        s[naxes] = slice(None)
        paxes = np.setxor1d(range(u.ndim), naxes)
        if u.ndim == 2:
            for i in range(u.shape[paxes[0]]):
                if i == 0 and is_rank_zero:
                    continue
                s[paxes[0]] = i
                s0 = tuple(s)
                sol(u[s0], data[i])

        elif u.ndim == 3:
            for i in range(u.shape[paxes[0]]):
                s[paxes[0]] = i
                for j in range(u.shape[paxes[1]]):
                    if i == 0 and j == 0 and is_rank_zero:
                        continue
                    s[paxes[1]] = j
                    s0 = tuple(s)
                    sol(u[s0], data[i, j])
        return u

    def fast_solve(self, u, b, solvers1D, naxes):
        if u is not b:
            s = tuple([slice(0, i) for i in u.shape])
            u[s] = b[s]
        # Solve first for the possibly different Fourier wavenumber 0, or (0, 0) in 3D
        # All other wavenumbers we assume have the same solver
        sol0 = solvers1D[0] if u.ndim == 2 else solvers1D[0][0]
        sol1 = solvers1D[-1] if u.ndim == 2 else solvers1D[-1][-1]
        is_rank_zero = comm.Get_rank() == 0

        if is_rank_zero:
            s = [0]*u.ndim
            s[naxes] = slice(None)
            s = tuple(s)
            sol0.inner_solve(u[s], sol0._inner_arg)

        data = self.get_data(is_rank_zero)
        sol = optimizer(sol1.inner_solve, False)
        u = self.solve_data(u, data, sol, naxes, is_rank_zero)

    def solve(self, u, b, solvers1D, naxes):
        if u is not b:
            s = tuple([slice(0, i) for i in u.shape])
            u[s] = b[s]

        s = [0]*u.ndim
        s[naxes] = slice(None)
        paxes = np.setxor1d(range(u.ndim), naxes)
        if u.ndim == 2:
            for i, sol in enumerate(solvers1D):
                s[paxes[0]] = i
                s0 = tuple(s)
                sol.inner_solve(u[s0], sol._inner_arg)

        elif u.ndim == 3:
            for i, m in enumerate(solvers1D):
                s[paxes[0]] = i
                for j, sol in enumerate(m):
                    s[paxes[1]] = j
                    s0 = tuple(s)
                    sol.inner_solve(u[s0], sol._inner_arg)

    def __call__(self, b, u=None, constraints=(), fast=True):
        """Solve problem with one non-diagonal direction

        Parameters
        ----------
        b : array, right hand side
        u : array, solution
        constraints : tuple of 2-tuples
            Each 2-tuple (row, value) is a constraint set for the non-periodic
            direction, for Fourier index 0 in 2D and (0, 0) in 3D
        fast : bool
            Use fast routine if possible. A fast routine is possible
            for any system of matrices with a tailored solver, like the
            TDMA, PDMA, FDMA and TwoDMA.

        """
        if u is None:
            u = b
        else:
            pass
            #assert u.shape == b.shape

        if len(self.bc_mats) > 0:
            u.set_boundary_dofs()
            w0 = np.zeros_like(b)
            for bc_mat in self.bc_mats:
                b -= bc_mat.matvec(u, w0)

        b = self.apply_constraints(b, constraints)

        if not self._lu:
            self.perform_lu()

        sol1 = self.solvers1D[-1] if u.ndim == 2 else self.solvers1D[-1][-1]
        if isinstance(sol1._inner_arg, tuple):
            fast = False

        if not fast:
            self.solve(u, b, self.solvers1D, self.naxes)
        else:
            self.fast_solve(u, b, self.solvers1D, self.naxes)

        if hasattr(u, 'set_boundary_dofs'):
            u.set_boundary_dofs()
        return u

class BlockMatrixSolver:
    def __init__(self, mats):
        assert isinstance(mats, (BlockMatrix, list))
        self.bc_mat = None
        self._lu = None
        if isinstance(mats, BlockMatrix):
            mats = mats.get_mats()
        bc_mats = extract_bc_matrices([mats])
        assert len(mats) > 0
        self.mat = BlockMatrix(mats)
        if len(bc_mats) > 0:
            self.bc_mat = BlockMatrix(bc_mats)

    @staticmethod
    def apply_constraint(A, b, offset, i, constraint):
        if constraint is None or comm.Get_rank() > 0:
            return A, b

        if isinstance(i, int):
            if i > 0:
                return A, b

        if isinstance(i, tuple):
            if np.sum(np.array(i)) > 0:
                return A, b

        row = offset + constraint[1]
        assert isinstance(constraint, tuple)
        assert len(constraint) == 3
        val = constraint[2]
        b[row] = val
        if A is not None:
            A = A.tolil()
            r = A.getrow(row).nonzero()
            A[(row, r[1])] = 0
            A[row, row] = 1
            A = A.tocsc()
        return A, b

    def __call__(self, b, u=None, constraints=()):
        from .forms.arguments import Function
        import scipy.sparse as sp
        space = b.function_space()
        if u is None:
            u = Function(space)
        else:
            assert u.shape == b.shape

        if self.bc_mat: # Add contribution to right hand side due to inhomogeneous boundary conditions
            u.set_boundary_dofs()
            w0 = np.zeros_like(b)
            b -= self.bc_mat.matvec(u, w0)

        nvars = b.shape[0] if len(b.shape) > space.dimensions else 1
        u = np.expand_dims(u, 0) if nvars == 1 else u
        b = np.expand_dims(b, 0) if nvars == 1 else b
        for con in constraints:
            assert len(con) == 3
            assert isinstance(con[0], Integral)
            assert isinstance(con[1], Integral)
            assert isinstance(con[2], Number)
        self.mat.assemble()
        if self._lu is None:
            self._lu = {}

        daxes = space.get_diagonal_axes()
        if len(daxes) == space.dimensions:
            # Only Fourier spaces, all diagonal
            assert len(daxes) == space.dimensions
            Ai = self.mat._Ai[0]
            gi = b.flatten()
            if isinstance(self._lu, dict):
                for con in constraints:
                    Ai, gi = self.apply_constraint(Ai, gi, np.sum(np.array(space.dims()[:con[0]])), 0, con)
            else:
                for con in constraints:
                    _, gi = self.apply_constraint(None, gi, np.sum(np.array(space.dims()[:con[0]])), 0, con)
            lu = sp.linalg.splu(Ai, permc_spec=config['matrix']['block']['permc_spec'])
            self._lu = lu
            u[:] = lu.solve(gi).reshape(u.shape)

        else:
            sl, dims = space._get_ndiag_slices_and_dims()
            gi = np.zeros(dims[-1], dtype=b.dtype)
            for key, Ai in self.mat._Ai.items():
                if len(daxes) > 0:
                    sl.T[daxes+1] = key if isinstance(key, int) else np.array(key)[:, None]
                gi = b.copy_to_flattened(gi, key, dims, sl)
                if key in self._lu:
                    lu = self._lu[key]
                    for con in constraints:
                        _, gi = self.apply_constraint(None, gi, dims[con[0]], key, con)
                else:
                    for con in constraints:
                        Ai, gi = self.apply_constraint(Ai, gi, dims[con[0]], key, con)

                    lu = sp.linalg.splu(Ai, permc_spec=config['matrix']['block']['permc_spec'])
                    self._lu[key] = lu

                if b.dtype.char in 'fdg' or lu.U.dtype.char in 'FDG':
                    u = u.copy_from_flattened(lu.solve(gi), key, dims, sl)
                else:
                    u.real = u.real.copy_from_flattened(lu.solve(gi.real), key, dims, sl)
                    u.imag = u.imag.copy_from_flattened(lu.solve(gi.imag), key, dims, sl)

        u = u.reshape(u.shape[1:]) if nvars == 1 else u
        b = b.reshape(b.shape[1:]) if nvars == 1 else b
        return u



import numpy as np
from numba import jit

__all__ = ['chebval']

def chebval(x, c):
    c = np.array(c, ndmin=1, copy=True)
    if c.dtype.char in '?bBhHiIlLqQpP':
        c = c.astype(np.double)
    if isinstance(x, int):
        x = float(x)
    if isinstance(x, (tuple, list)):
        x = np.asarray(x)
    y = np.zeros_like(x, dtype=c.dtype)
    c0 = np.zeros_like(x, dtype=c.dtype)
    c1 = np.zeros_like(x, dtype=c.dtype)
    tmp = np.zeros_like(x, dtype=c.dtype)
    x2 = np.zeros_like(x)
    y = _chebval(x, c, y, c0, c1, tmp, x2)
    return y

@jit
def _chebval(x, c, y, c0, c1, tmp, x2):
    N = c.shape[0]
    M = x.shape[0]
    if N == 1:
        c0[:] = c[0]
        c1[:] = 0
    elif N == 2:
        c0[:] = c[0]
        c1[:] = c[1]
    else:
        for j in range(M):
            x2[j] = 2*x[j]
            c0[j] = c[-2]
            c1[j] = c[-1]
        for i in range(3, N + 1):
            for j in range(M):
                tmp[j] = c0[j]
                c0[j] = c[-i] - c1[j]
                c1[j] = tmp[j] + c1[j]*x2[j]
    for j in range(M):
        y[j] = c0[j] + x[j]*c1[j]
    return y



import numbers
import sympy as sp
import numpy as np
from shenfun.config import config

class Coordinates:
    """Class for handling curvilinear coordinates

    Parameters
    ----------
    psi : tuple or sp.Symbol
        The new coordinates
    rv : tuple
        The position vector in terms of the new coordinates
    assumptions : Sympy assumptions
        One or more `Sympy assumptions <https://docs.sympy.org/latest/modules/assumptions/index.html>`_
    replace : sequence of two-tuples
        Use Sympy's replace with these two-tuples
    measure : Python function to replace Sympy's count_ops.
        For example, to discourage the use of powers in an expression use::

        def discourage_powers(expr):
            POW = sp.Symbol('POW')
            count = sp.count_ops(expr, visual=True)
            count = count.replace(POW, 100)
            count = count.replace(sp.Symbol, type(sp.S.One))
            return count
    """
    def __init__(self, psi, rv, assumptions=True, replace=(), measure=sp.count_ops):
        self._psi = (psi,) if isinstance(psi, sp.Symbol) else psi
        self._rv = rv
        self._assumptions = assumptions
        self._replace = replace
        self._measure = measure
        self._hi = None
        self._b = None
        self._bt = None
        self._e = None
        self._g = None
        self._gt = None
        self._gn = None
        self._ct = None
        self._det_g = {True: None, False: None}
        self._sqrt_det_g = {True: None, False: None}

    @property
    def b(self):
        return self.get_covariant_basis()

    @property
    def e(self):
        return self.get_normal_basis()

    @property
    def hi(self):
        return self.get_scaling_factors()

    @property
    def sg(self):
        if self.is_cartesian:
            return 1
        return self.get_sqrt_det_g(True)

    @property
    def coordinates(self):
        return (self.psi, self.rv, self._assumptions, self._replace)

    @property
    def psi(self):
        return self._psi

    @property
    def rv(self):
        return self._rv

    @property
    def is_orthogonal(self):
        return sp.Matrix(self.get_covariant_metric_tensor()).is_diagonal()

    @property
    def is_cartesian(self):
        if len(self.psi) != len(self.rv):
            return False
        return sp.Matrix(self.get_covariant_metric_tensor()).is_Identity

    def get_det_g(self, covariant=True):
        """Return determinant of covariant metric tensor"""
        if self._det_g[covariant] is not None:
            return self._det_g[covariant]
        if covariant:
            g = sp.Matrix(self.get_covariant_metric_tensor()).det()
        else:
            g = sp.Matrix(self.get_contravariant_metric_tensor()).det()
        g = g.factor()
        g = self.refine(g)
        g = sp.simplify(g, measure=self._measure)
        self._det_g[covariant] = g
        return g

    def get_sqrt_det_g(self, covariant=True):
        """Return square root of determinant of covariant metric tensor"""
        if self._sqrt_det_g[covariant] is not None:
            return self._sqrt_det_g[covariant]
        g = self.get_det_g(covariant)
        sg = sp.simplify(sp.sqrt(g), measure=self._measure)
        sg = self.refine(sg)
        if isinstance(sg, numbers.Number):
            if isinstance(sg, numbers.Real):
                sg = float(sg)
            elif isinstance(sg, numbers.Complex):
                sg = complex(sg)
            else:
                raise NotImplementedError

        self._sqrt_det_g[covariant] = sg
        return sg

    def get_cartesian_basis(self):
        """Return Cartesian basis vectors"""
        return np.eye(len(self.rv), dtype=object)

    def get_scaling_factors(self):
        """Return scaling factors"""
        if self._hi is not None:
            return self._hi
        hi = np.zeros_like(self.psi)

        for i, s in enumerate(np.sum(self.b**2, axis=1)):
            hi[i] = sp.sqrt(self.refine(sp.simplify(s, measure=self._measure)))
            hi[i] = self.refine(hi[i])

        self._hi = hi
        return hi

    def get_normal_basis(self):
        if self._e is not None:
            return self._e
        b = self.b
        e = np.zeros_like(b)
        for i, bi in enumerate(b):
            l = sp.sqrt(sp.simplify(np.dot(bi, bi)))
            l = self.refine(l)
            e[i] = bi / l
        self._e = e
        return e

    def get_covariant_basis(self):
        """Return covariant basisvectors"""
        if self._b is not None:
            return self._b
        b = np.zeros((len(self.psi), len(self.rv)), dtype=object)
        for i, ti in enumerate(self.psi):
            for j, rj in enumerate(self.rv):
                b[i, j] = rj.diff(ti, 1)
                b[i, j] = sp.simplify(b[i, j], measure=self._measure)
                b[i, j] = self.refine(b[i, j])

        #if len(self.psi) == 2 and len(self.rv) == 3:
        #    b[-1] = np.cross(b[0], b[1])
        #    bl = self.refine(sp.sqrt(sp.simplify(np.dot(b[-1], b[-1]))))
        #    b[-1] = b[-1] / bl
        #    for j in range(len(self.rv)):
        #        b[-1, j] = sp.simplify(b[-1, j], measure=self._measure)
        #        b[-1, j] = self.refine(b[-1, j])
        self._b = b
        return b

    def get_contravariant_basis(self):
        """Return contravariant basisvectors"""
        if self._bt is not None:
            return self._bt
        bt = np.zeros_like(self.b)
        g = self.get_contravariant_metric_tensor()
        b = self.b
        for i in range(len(self.psi)):
            for j in range(len(self.psi)):
                bt[i] += g[i, j]*b[j]

        for i in range(len(self.psi)):
            for j in range(len(self.psi)):
                bt[i, j] = sp.simplify(bt[i, j], measure=self._measure)
        self._bt = bt
        return bt

    def get_normal_metric_tensor(self):
        """Return normal metric tensor"""
        if self._gn is not None:
            return self._gn
        gn = np.zeros((len(self.psi), len(self.psi)), dtype=object)
        e = self.e
        for i in range(len(self.psi)):
            for j in range(len(self.psi)):
                gn[i, j] = sp.simplify(np.dot(e[i], e[j]).expand(), measure=self._measure)
                gn[i, j] = self.refine(gn[i, j])

        self._gn = gn
        return gn

    def get_covariant_metric_tensor(self):
        """Return covariant metric tensor"""
        if self._g is not None:
            return self._g
        g = np.zeros((len(self.psi), len(self.psi)), dtype=object)
        b = self.b
        for i in range(len(self.psi)):
            for j in range(len(self.psi)):
                g[i, j] = sp.simplify(np.dot(b[i], b[j]).expand(), measure=self._measure)
                g[i, j] = self.refine(g[i, j])

        self._g = g
        return g

    def get_contravariant_metric_tensor(self):
        """Return contravariant metric tensor"""
        if self._gt is not None:
            return self._gt
        g = self.get_covariant_metric_tensor()
        gt = sp.Matrix(g).inv()
        for i in range(gt.shape[0]):
            for j in range(gt.shape[1]):
                gt[i, j] = sp.simplify(gt[i, j], measure=self._measure)
        gt = np.array(gt)
        self._gt = gt
        return gt

    def get_christoffel_second(self):
        """Return Christoffel symbol of second kind"""
        if self._ct is not None:
            return self._ct
        b = self.get_covariant_basis()
        bt = self.get_contravariant_basis()
        ct = np.zeros((len(self.psi),)*3, object)
        for i in range(len(self.psi)):
            for j in range(len(self.psi)):
                for k in range(len(self.psi)):
                    ct[k, i, j] = sp.simplify(np.dot(np.array([bij.diff(self.psi[j], 1) for bij in b[i]]), bt[k]), measure=self._measure)
        self._ct = ct
        return ct

    def get_metric_tensor(self, kind='normal'):
        if kind == 'covariant':
            gij = self.get_covariant_metric_tensor()
        elif kind == 'contravariant':
            gij = self.get_contravariant_metric_tensor()
        elif kind == 'normal':
            gij = self.get_normal_metric_tensor()
        else:
            raise NotImplementedError
        return gij

    def get_basis(self, kind='normal'):
        if kind == 'covariant':
            return self.get_covariant_basis()
        assert kind == 'normal'
        return self.get_normal_basis()

    def refine(self, sc):
        sc = sp.refine(sc, self._assumptions)
        for a, b in self._replace:
            sc = sc.replace(a, b)
        return sc

    def subs(self, s0, s1):
        b = self.get_covariant_basis()
        for i in range(b.shape[0]):
            for j in range(b.shape[1]):
                b[i, j] = b[i, j].subs(s0, s1)

        g = self.get_covariant_metric_tensor()
        gt = self.get_contravariant_metric_tensor()
        for i in range(g.shape[0]):
            for j in range(g.shape[1]):
                g[i, j] = g[i, j].subs(s0, s1)
                gt[i, j] = gt[i, j].subs(s0, s1)

        sg = self.get_sqrt_det_g().subs(s0, s1)
        self._sqrt_det_g[True] = sg

        hi = self.get_scaling_factors()
        for i in range(len(hi)):
            hi[i] = hi[i].subs(s0, s1)

        self._psi = tuple([p.subs(s0, s1) for p in self._psi])
        self._rv = tuple([r.subs(s0, s1) for r in self._rv])

    def latex_basis_vectors(self, symbol_names=None, replace=None, kind=None):
        if kind is None:
            kind = config['basisvectors']
        if kind == 'covariant':
            b = self.get_covariant_basis()
        elif kind == 'contravariant':
            b = self.get_contravariant_basis()
        else:
            b = self.get_normal_basis()
        psi = self.psi
        symbols = {p: str(p) for p in psi}
        if symbol_names is not None:
            symbols = symbol_names

        k = {0: '\\mathbf{i}', 1: '\\mathbf{j}', 2: '\\mathbf{k}'}
        m = ' '
        bl = 'e' if kind == 'normal' else 'b'
        for i, p in enumerate(psi):
            if kind in ('covariant', 'normal'):
                m += '\\mathbf{%s}_{%s} ='%(bl, symbols[p])
            else:
                m += '\\mathbf{%s}^{%s} ='%(bl, symbols[p])
            for j in range(b.shape[1]):
                if b[i, j] == 1:
                    m += (k[j]+'+')
                elif b[i, j] != 0:
                    if replace is not None:
                        for repl in replace:
                            assert len(repl) == 2
                            b[i, j] = b[i, j].replace(*repl)
                    sl = sp.latex(b[i, j], symbol_names=symbols)
                    if sl.startswith('-') and not isinstance(b[i, j], sp.Add):
                        m = m.rstrip('+')
                    if isinstance(b[i, j], sp.Add):
                        sl = '\\left(%s\\right)'%(sl)
                    m += (sl+'\\,'+k[j]+'+')

            m = m.rstrip('+')
            m += ' \\\\ '
        m += ' '
        return r'%s'%(m)



#pylint: disable=missing-docstring

import numpy as np
from .bases import *
from .matrices import *

def energy_fourier(u, T):
    r"""Compute the energy of u using Parceval's theorem

    .. math::

        \int abs(u)^2 dx = N*\sum abs(u_hat)^2

    Parameters
    ----------
        u : Array
            The Fourier coefficients
        T : TensorProductSpace

    See https://en.wikipedia.org/wiki/Parseval's_theorem

    """

    if not hasattr(T, 'comm'):
        # Just a 1D basis
        assert u.ndim == 1
        if isinstance(T, R2C):
            if u.shape[0] % 2 == 0:
                result = (2*np.sum(abs(u[1:-1])**2) +
                          np.sum(abs(u[0])**2) +
                          np.sum(abs(u[-1])**2))
            else:
                result = (2*np.sum(abs(u[1:])**2) +
                          np.sum(abs(u[0])**2))

        else:
            result = np.sum(abs(u)**2)
        return result

    comm = T.comm
    assert np.all([base.family() == 'fourier' for base in T.bases])
    real = False
    for axis, base in enumerate(T.bases):
        if isinstance(base, R2C):
            real = True
            break

    if real:
        s = [slice(None)]*u.ndim
        uaxis = axis + u.ndim-len(T.bases)
        if T.forward.output_pencil.subcomm[axis].Get_size() == 1:
            # aligned in r2c direction
            if base.N % 2 == 0:
                s[uaxis] = slice(1, -1)
                result = 2*np.sum(abs(u[tuple(s)])**2)
                s[uaxis] = 0
                result += np.sum(abs(u[tuple(s)])**2)
                s[uaxis] = -1
                result += np.sum(abs(u[tuple(s)])**2)
            else:
                s[uaxis] = slice(1, None)
                result = 2*np.sum(abs(u[tuple(s)])**2)
                s[uaxis] = 0
                result += np.sum(abs(u[tuple(s)])**2)

        else:
            # Data not aligned along r2c axis. Need to check about 0 and -1
            if base.N % 2 == 0:
                s[uaxis] = slice(1, -1)
                result = 2*np.sum(abs(u[tuple(s)])**2)
                s[uaxis] = 0
                if T.local_slice(True)[axis].start == 0:
                    result += np.sum(abs(u[tuple(s)])**2)
                else:
                    result += 2*np.sum(abs(u[tuple(s)])**2)
                s[uaxis] = -1
                if T.local_slice(True)[axis].stop == T.dims()[axis]:
                    result += np.sum(abs(u[tuple(s)])**2)
                else:
                    result += 2*np.sum(abs(u[tuple(s)])**2)
            else:
                s[uaxis] = slice(1, None)
                result = 2*np.sum(abs(u[tuple(s)])**2)
                s[uaxis] = 0
                if T.local_slice(True)[axis].start == 0:
                    result += np.sum(abs(u[tuple(s)])**2)
                else:
                    result += 2*np.sum(abs(u[tuple(s)])**2)

    else:
        result = np.sum(abs(u[...])**2)

    result = comm.allreduce(result)
    return result



r"""
Module for some integrators.

Integrators are set up to solve equations like

.. math::

    \frac{\partial u}{\partial t} = L u + N(u)

where :math:`u` is the solution, :math:`L` is a linear operator and
:math:`N(u)` is the nonlinear part of the right hand side.

There are two kinds of integrators, or time steppers. The first are
ment to be subclassed and used one at the time. These are

    - IRK3:     Implicit third order Runge-Kutta
    - RK4:      Runge-Kutta fourth order
    - ETD:      Exponential time differencing Euler method
    - ETDRK4:   Exponential time differencing Runge-Kutta fourth order

See, e.g.,
H. Montanelli and N. Bootland "Solving periodic semilinear PDEs in 1D, 2D and
3D with exponential integrators", https://arxiv.org/pdf/1604.08900.pdf

The second kind is ment to be used for systems of equations, one class
instance for each equation. These are mainly IMEX Runge Kutta integrators:

    - IMEXRK3
    - IMEXRK111
    - IMEXRK222
    - IMEXRK443

See, e.g., https://github.com/spectralDNS/shenfun/blob/master/demo/ChannelFlow
for an example of use.

Note
----
`RK4`, `ETD` and `ETDRK4` can only be used with Fourier function spaces,
as they assume all matrices are diagonal.

"""
import copy
import types
import numpy as np
from shenfun import Function, TPMatrix, TrialFunction, TestFunction,\
    inner, la, Expr, CompositeSpace, BlockMatrix, SparseMatrix, \
    get_simplified_tpmatrices, ScipyMatrix, Inner

__all__ = ('IRK3', 'BackwardEuler', 'RK4', 'ETDRK4', 'ETD',
           'IMEXRK3', 'IMEXRK111', 'IMEXRK222', 'IMEXRK443')

#pylint: disable=unused-variable

class IntegratorBase:
    """Abstract base class for integrators

    Parameters
    ----------
        T : TensorProductSpace
        L : function
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
                 To be called at the end of a timestep
        params : dictionary
                 Any relevant keyword arguments

    """

    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        _p = {'dt': 0}
        _p.update(params)
        self.params = _p
        self.T = T
        if L is not None:
            self.LinearRHS = types.MethodType(L, self)
        if N is not None:
            self.NonlinearRHS = types.MethodType(N, self)
        if update is not None:
            self.update = types.MethodType(update, self)

    def update(self, u, u_hat, t, tstep, **par):
        pass

    def LinearRHS(self, *args, **kwargs):
        return 0

    def NonlinearRHS(self, *args, **kwargs):
        return 0

    def setup(self, dt):
        """Set up solver"""
        pass

    def solve(self, u, u_hat, dt, trange):
        """Integrate forward in time

        Parameters
        ----------
            u : array
                The solution array in physical space
            u_hat : array
                The solution array in spectral space
            dt : float
                Timestep
            trange : two-tuple
                Time and end time
        """
        pass


class IRK3(IntegratorBase):
    """Third order implicit Runge Kutta

    Parameters
    ----------
        T : TensorProductSpace
        L : function of TrialFunction(T)
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
            To be called at the end of a timestep
        params : dictionary
            Any relevant keyword arguments

    """
    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        IntegratorBase.__init__(self, T, L=L, N=N, update=update, **params)
        self.dU = Function(T)
        self.dU1 = Function(T)
        self.a = (8./15., 5./12., 3./4.)
        self.b = (0.0, -17./60., -5./12.)
        self.c = (0.0, 8./15., 2./3., 1)
        self.solver = None
        self.bm = None
        self.rhs_mats = None
        self.w0 = Function(self.T)
        self.mask = None
        if hasattr(T, 'get_mask_nyquist'):
            self.mask = T.get_mask_nyquist()

    def setup(self, dt):
        if isinstance(self.T, CompositeSpace):
            assert self.T.tensor_rank > 0, 'IRK3 only works for tensors, not generic CompositeSpaces'

        self.params['dt'] = dt
        u = TrialFunction(self.T)
        v = TestFunction(self.T)

        # Note that we are here assembling implicit left hand side matrices,
        # as well as matrices that can be used to assemble the right hand side
        # much faster through matrix-vector products

        a, b = self.a, self.b
        self.solver = []
        self.rhs_mats = []
        u0 = self.LinearRHS(u)
        for rk in range(3):
            if u0:
                mats = inner(v, u-((a[rk]+b[rk])*dt/2)*u0)
            else:
                mats = inner(v, u)
            if self.T.dimensions == 1:
                self.solver.append(la.Solver(mats))

            elif self.T.tensor_rank == 0:
                if len(mats[0].naxes) == 1:
                    self.solver.append(la.SolverGeneric1ND(mats))
                elif len(mats[0].naxes) == 2:
                    self.solver.append(la.SolverGeneric2ND(mats))
                else:
                    raise NotImplementedError
            else:
                self.solver.append(la.BlockMatrixSolver(mats))

            if u0:
                rhs_mats = inner(v, u+((a[rk]+b[rk])*dt/2)*u0)
            else:
                rhs_mats = inner(v, u)
            mat = ScipyMatrix if self.T.dimensions == 1 else BlockMatrix
            self.rhs_mats.append(mat(rhs_mats))

    def compute_rhs(self, u, u_hat, dU, dU1, rk):
        a = self.a[rk]
        b = self.b[rk]
        dt = self.params['dt']
        dU = self.NonlinearRHS(u, u_hat, dU, **self.params)
        if self.mask is not None:
            dU.mask_nyquist(self.mask)
        w1 = dU*a*dt + dU1*b*dt
        dU1[:] = dU
        if isinstance(dU, np.ndarray):
            dU[:] = w1
        return dU

    def solve(self, u, u_hat, dt, trange):
        if self.solver is None or abs(self.params['dt']-dt) > 1e-12:
            self.setup(dt)
        t, end_time = trange
        tstep = self.tstep = 0
        while t < end_time-1e-8:
            self.tstep = tstep
            for rk in range(3):
                self.params['ti'] = t+self.c[rk]*dt
                dU = self.compute_rhs(u, u_hat, self.dU, self.dU1, rk)
                dU += self.rhs_mats[rk].matvec(u_hat, self.w0)
                u_hat = self.solver[rk](dU, u=u_hat)
                if self.mask is not None:
                    u_hat.mask_nyquist(self.mask)

            t += dt
            tstep += 1
            self.update(u, u_hat, t, tstep, **self.params)
        return u_hat

class BackwardEuler(IntegratorBase):
    """First order backward Euler

    Parameters
    ----------
        T : TensorProductSpace
        L : function of TrialFunction(T)
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
            To be called at the end of a timestep
        params : dictionary
            Any relevant keyword arguments

    """
    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        IntegratorBase.__init__(self, T, L=L, N=N, update=update, **params)
        self.dU = Function(T)
        self.dU1 = Function(T)
        self.solver = None
        self.rhs_mats = None
        self.w0 = Function(self.T)
        self.mask = None
        if hasattr(T, 'get_mask_nyquist'):
            self.mask = T.get_mask_nyquist()

    def setup(self, dt):
        if isinstance(self.T, CompositeSpace):
            assert self.T.tensor_rank > 0, 'BackwardEuler only works for tensors, not generic CompositeSpaces'

        self.params['dt'] = dt
        u = TrialFunction(self.T)
        v = TestFunction(self.T)
        mats = inner(u-dt*self.LinearRHS(u), v)
        M = inner(u, v)

        if self.T.dimensions == 1:
            self.solver = la.Solve(mats)
            self.rhs_mats = M
            return

        if self.T.tensor_rank == 0:
            if len(mats[0].naxes) == 1:
                self.solver = la.SolverGeneric1ND(mats)
            elif len(mats[0].naxes) == 2:
                self.solver = la.SolverGeneric2ND(mats)
            else:
                raise NotImplementedError
        else:
            self.solver = la.BlockMatrixSolver(mats)
        self.rhs_mats = BlockMatrix(M if isinstance(M, list) else [M])

    def compute_rhs(self, u, u_hat, dU, dU1):
        dt = self.params['dt']
        dU = self.NonlinearRHS(u, u_hat, dU, **self.params)
        if self.mask:
            dU.mask_nyquist(self.mask)
        w1 = dU*2*dt - dU1*dt
        dU1[:] = dU
        return w1

    def solve(self, u, u_hat, dt, trange):
        if self.solver is None or abs(self.params['dt']-dt) > 1e-12:
            self.setup(dt)
        t, end_time = trange
        tstep = self.tstep = 0
        while t < end_time-1e-8:
            self.params['ti'] = t
            self.tstep = tstep
            dU = self.compute_rhs(u, u_hat, self.dU, self.dU1)
            dU += self.rhs_mats.matvec(u_hat, self.w0)
            u_hat = self.solver(dU, u=u_hat)
            if self.mask:
                u_hat.mask_nyquist(self.mask)
            t += dt
            tstep += 1
            self.update(u, u_hat, t, tstep, **self.params)
        return u_hat


class ETD(IntegratorBase):
    """Exponential time differencing Euler method

    H. Montanelli and N. Bootland "Solving periodic semilinear PDEs in 1D, 2D and
    3D with exponential integrators", https://arxiv.org/pdf/1604.08900.pdf

    Parameters
    ----------
        T : TensorProductSpace
        L : function
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
            To be called at the end of a timestep
        params : dictionary
            Any relevant keyword arguments
    """

    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        IntegratorBase.__init__(self, T, L=L, N=N, update=update, **params)
        self.dU = Function(T)
        self.psi = None
        self.ehL = None

    def setup(self, dt):
        """Set up ETD ODE solver"""
        self.params['dt'] = dt
        u = TrialFunction(self.T)
        v = TestFunction(self.T)
        L = self.LinearRHS(u, **self.params)
        if isinstance(L, Expr):
            L = inner(v, L)
            L = get_simplified_tpmatrices(L)[0]
        if isinstance(L, list):
            assert self.T.tensor_rank == 1
            assert L[0].isidentity()
            L = L[0].scale
            # Use only L[0] and let numpy broadcasting take care of the rest
        elif isinstance(L, TPMatrix):
            assert L.isidentity()
            L = L.scale
        elif isinstance(L, SparseMatrix):
            L.simplify_diagonal_matrices()
            L = L.scale

        L = np.atleast_1d(L)
        hL = L*dt
        self.ehL = np.exp(hL)
        M = 50
        psi = self.psi = np.zeros(hL.shape, dtype=float)
        for k in range(1, M+1):
            ll = hL+np.exp(np.pi*1j*(k-0.5)/M)
            psi += ((np.exp(ll)-1.)/ll).real
        psi /= M

    def solve(self, u, u_hat, dt, trange):
        """Integrate forward in time

        Parameters
        ----------
            u : array
                The solution array in physical space
            u_hat : array
                The solution array in spectral space
            dt : float
                Timestep
            trange : two-tuple
                Time and end time
        """
        if self.psi is None or abs(self.params['dt']-dt) > 1e-12:
            self.setup(dt)
        t, end_time = trange
        tstep = 0
        while t < end_time-1e-8:
            t += dt
            tstep += 1
            self.dU = self.NonlinearRHS(u, u_hat, self.dU, **self.params)
            u_hat[:] = self.ehL*u_hat + dt*self.psi*self.dU
            self.update(u, u_hat, t, tstep, **self.params)
        return u_hat


class ETDRK4(IntegratorBase):
    """Exponential time differencing Runge-Kutta 4'th order method

    H. Montanelli and N. Bootland "Solving periodic semilinear PDEs in 1D, 2D and
    3D with exponential integrators", https://arxiv.org/pdf/1604.08900.pdf

    Parameters
    ----------
        T : TensorProductSpace
        L : function
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
            To be called at the end of a timestep
        params : dictionary
            Any relevant keyword arguments
    """
    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        IntegratorBase.__init__(self, T, L=L, N=N, update=update, **params)
        self.U_hat0 = Function(T)
        self.U_hat1 = Function(T)
        self.dU = Function(T)
        self.dU0 = Function(T)
        self.V2 = Function(T)
        self.psi = np.zeros((4,)+self.U_hat0.shape, dtype=float)
        self.a = None
        self.b = [0.5, 0.5, 0.5]
        self.ehL = None
        self.ehL_h = None

    def setup(self, dt):
        """Set up ETDRK4 ODE solver"""
        self.params['dt'] = dt
        u = TrialFunction(self.T)
        v = TestFunction(self.T)
        L = self.LinearRHS(u, **self.params)
        if isinstance(L, Expr):
            L = inner(v, L)
            if isinstance(L, list):
                L = get_simplified_tpmatrices(L)[0]
        if isinstance(L, list):
            assert self.T.tensor_rank == 1
            assert L[0].isidentity()
            L = L[0].scale
            # Use only L[0] and let numpy broadcasting take care of the rest
        elif isinstance(L, TPMatrix):
            assert L.isidentity()
            L = L.scale
        elif isinstance(L, SparseMatrix):
            L.simplify_diagonal_matrices()
            L = L.scale

        L = np.atleast_1d(L)
        hL = L*dt
        self.ehL = np.exp(hL)
        self.ehL_h = np.exp(hL/2.)
        M = 50
        psi = self.psi = np.zeros((4,) + hL.shape, dtype=float)
        for k in range(1, M+1):
            ll = hL+np.exp(np.pi*1j*(k-0.5)/M)
            psi[0] += ((np.exp(ll)-1.)/ll).real
            psi[1] += ((np.exp(ll)-ll-1.)/ll**2).real
            psi[2] += ((np.exp(ll)-0.5*ll**2-ll-1.)/ll**3).real
            ll2 = hL/2.+np.exp(np.pi*1j*(k-0.5)/M)
            psi[3] += ((np.exp(ll2)-1.)/(ll2)).real

        psi /= M
        a = [psi[0]-3*psi[1]+4*psi[2]]
        a.append(2*psi[1]-4*psi[2])
        a.append(2*psi[1]-4*psi[2])
        a.append(-psi[1]+4*psi[2])
        self.a = a

    def solve(self, u, u_hat, dt, trange):
        """Integrate forward in time

        Parameters
        ----------
            u : array
                The solution array in physical space
            u_hat : array
                The solution array in spectral space
            dt : float
                Timestep
            trange : two-tuple
                Time and end time
        """
        if self.a is None or abs(self.params['dt']-dt) > 1e-12:
            self.setup(dt)
        t, end_time = trange
        tstep = 0
        while t < end_time-1e-8:
            t += dt
            tstep += 1

            self.U_hat0[:] = u_hat*self.ehL_h
            self.U_hat1[:] = u_hat*self.ehL
            for rk in range(4):
                self.dU = self.NonlinearRHS(u, u_hat, self.dU, **self.params)
                if rk < 2:
                    u_hat[:] = self.U_hat0 + self.b[rk]*dt*self.psi[3]*self.dU
                elif rk == 2:
                    u_hat[:] = self.ehL_h*self.V2 + self.b[rk]*dt*self.psi[3]*(2*self.dU-self.dU0)

                if rk == 0:
                    self.dU0[:] = self.dU
                    self.V2[:] = u_hat

                self.U_hat1 += self.a[rk]*dt*self.dU
            u_hat[:] = self.U_hat1
            self.update(u, u_hat, t, tstep, **self.params)
        return u_hat


class RK4(IntegratorBase):
    """Regular 4'th order Runge-Kutta integrator

    Parameters
    ----------
        T : TensorProductSpace
        L : function
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
            To be called at the end of a timestep
        params : dictionary
            Any relevant keyword arguments
    """
    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        IntegratorBase.__init__(self, T, L=L, N=N, update=update, **params)
        self.U_hat0 = Function(T)
        self.U_hat1 = Function(T)
        self.dU = Function(T)
        self.a = np.array([1./6., 1./3., 1./3., 1./6.])
        self.b = np.array([0.5, 0.5, 1.])

    def setup(self, dt):
        """Set up RK4 ODE solver"""
        self.params['dt'] = dt

    def solve(self, u, u_hat, dt, trange):
        """Integrate forward in end_time

        Parameters
        ----------
            u : array
                The solution array in physical space
            u_hat : array
                The solution array in spectral space
            dt : float
                Timestep
            trange : two-tuple
                Time and end time
        """
        if self.a is None or abs(self.params['dt']-dt) > 1e-12:
            self.setup(dt)
        t, end_time = trange
        tstep = 0
        ut = TrialFunction(self.T)
        vt = TestFunction(self.T)
        L = self.LinearRHS(ut, **self.params)
        if isinstance(L, Expr):
            L = inner(vt, L)
            L = get_simplified_tpmatrices(L)[0]
        if isinstance(L, list):
            assert self.T.tensor_rank == 1
            assert L[0].isidentity()
            L = L[0].scale
            # Use only L[0] and let numpy broadcasting take care of the rest
        elif isinstance(L, TPMatrix):
            assert L.isidentity()
            L = L.scale
        elif isinstance(L, SparseMatrix):
            L.simplify_diagonal_matrices()
            L = L.scale

        while t < end_time-1e-8:
            t += dt
            tstep += 1
            self.U_hat0[:] = self.U_hat1[:] = u_hat
            for rk in range(4):
                dU = self.NonlinearRHS(u, u_hat, self.dU, **self.params)
                if isinstance(L, np.ndarray):
                    dU += L*u_hat
                if rk < 3:
                    u_hat[:] = self.U_hat0 + self.b[rk]*dt*dU
                self.U_hat1 += self.a[rk]*dt*dU
            u_hat[:] = self. U_hat1
            self.update(u, u_hat, t, tstep, **self.params)
        return u_hat


class IMEXRK3:
    r"""Solve partial differential equations of the form

    .. math::

        \frac{\partial u}{\partial t} = N+Lu, \quad (1)

    where :math:`N` is a nonlinear term and :math:`L` is a linear operator.

    This is a third order accurate implicit Runge-Kutta solver

    Parameters
    ----------
    v : :class:`.TestFunction`
    u : :class:`.Expr` or :class:`.Function`
        Representing :math:`u` in (1)
        If an :class:`.Expr`, then its basis must be a :class:`.Function`
        The :class:`.Function` will then hold the solution.
        If :math:`u` is a :class:`.Function`, then this will hold the solution.
    L : Linear operator
        Operates on :math:`u`
    N : :class:`.Expr` or sequence of :class:`.Expr`
        Nonlinear terms
    dt : number
        Time step
    solver : Linear solver, optional
    name : str, optional
    """
    def __init__(self, v, u, L, N, dt, solver=None, name='U-equation', latex=None):
        self.v = v
        self.u = u if isinstance(u, Expr) else Expr(u)
        self.u_ = self.u.basis()
        self.L = L
        self.N = N
        self.dt = dt
        T = self.T = v.function_space()
        self._solver = solver
        if solver is None:
            if v.dimensions == 1:
                self._solver = la.Solver
            elif len(T.get_nondiagonal_axes().ndim) == 1:
                self._solver = la.SolverGeneric1ND
            elif len(T.get_nondiagonal_axes().ndim) == 2:
               self._solver = la.SolverGeneric2ND
            else:
                raise NotImplementedError

        self.solvers = []
        self.linear_rhs = []
        self.nonlinear_rhs = None
        self.name = name
        self.latex = latex
        W = CompositeSpace([T, T])
        self.rhs = Function(W).v
        self.mask = None
        if hasattr(T, 'get_mask_nyquist'):
            self.mask = T.get_mask_nyquist()

    @classmethod
    def steps(cls):
        return 3

    def stages(self):
        a = (8./15., 5./12., 3./4.)
        b = (0.0, -17./60., -5./12.)
        c = (0.0, 8./15., 2./3., 1)
        return a, b, c

    def assemble(self):
        a, b, _ = self.stages()
        dt = self.dt
        ul = copy.copy(self.u)
        ul._basis = TrialFunction(self.u.function_space())
        L1 = self.L(ul)
        L2 = self.L(self.u)
        for rk in range(len(a)):
            mats = inner(self.v, ul - (a[rk]+b[rk])*dt/2*L1)
            self.solvers.append(self._solver(mats))
            self.linear_rhs.append(Inner(self.v, self.u + (a[rk]+b[rk])*dt/2*L2))
        if isinstance(self.N, (Expr, Function)):
            self.nonlinear_rhs = Inner(self.v, self.N)
        elif isinstance(self.N, list):
            self.nonlinear_rhs = sum([Inner(self.v, f) for f in self.N[1:]], start=Inner(self.v, self.N[0]))
        else:
            raise RuntimeError('Wrong type of nonlinear expression')

    def compute_rhs(self, rk=0):
        a, b, _ = self.stages()
        w0 = self.nonlinear_rhs()
        self.rhs[1] = self.dt*(a[rk]*w0+b[rk]*self.rhs[0])
        self.rhs[1] += self.linear_rhs[rk]()
        self.rhs[0] = w0
        if self.mask is not None:
            self.T.mask_nyquist(self.rhs[1], self.mask)
        return self.rhs

    def solve_step(self, rk):
        return self.solvers[rk](self.rhs[-1], self.u_)

class PDEIMEXRK:
    r"""Solve partial differential equations of the form

    .. math::

        \frac{\partial u}{\partial t} = N+Lu, \quad (1)

    where :math:`N` is a nonlinear term and :math:`L` is a linear operator.

    The solvers in this subclass are taken from::

        Ascher, Ruuth and Spiteri 'Implicit-explicit Runge-Kutta methods for
        time-dependent partial differential equations' Applied Numerical
        Mathematics, 25 (1997) 151-167

    Note in particular that we only use solvers satisfying condition
    (2.3).

    Parameters
    ----------
    v : :class:`.TestFunction`
    u : :class:`.Expr` or :class:`.Function`
        Representing :math:`u` in (1)
        If an :class:`.Expr`, then its basis must be a :class:`.Function`
        The :class:`.Function` will then hold the solution.
        If :math:`u` is a :class:`.Function`, then this will hold the solution.
    L : Linear operator
        Operates on :math:`u`
    N : :class:`.Expr` or sequence of :class:`.Expr`
        Nonlinear terms
    dt : number
        Time step
    solver : Linear solver, optional
    name : str, optional
    latex : str, optional
        optional representation of the equation to solve
    """
    def __init__(self, v, u, L, N, dt, solver=None, name='U-equation', latex=None):
        self.v = v
        self.u = u if isinstance(u, Expr) else Expr(u)
        self.u_ = self.u.basis()
        self.L = L
        self.N = N
        self.dt = dt
        self._solver = solver
        if solver is None:
            if v.dimensions > 1:
                self._solver = la.SolverGeneric1ND
            else:
                self._solver = la.Solver

        self.solvers = []
        self.linear_rhs = []
        self.nonlinear_rhs = None
        self.name = name
        self.latex = latex
        T = self.T = v.function_space()
        self.rhs = Function(T).v
        W = CompositeSpace([T]*self.steps())
        self.Krhs = Function(W).v # nonlinear terms
        if self.steps() == 1 and self.Krhs.ndim == 1:
            self.Krhs = np.expand_dims(self.Krhs, 0)

        if self.steps() > 1:
            WL = CompositeSpace([T]*(self.steps()-1))
            self.Lrhs = Function(WL).v # linear terms
            if self.steps() == 2 and self.Lrhs.ndim == 1:
                self.Lrhs = np.expand_dims(self.Lrhs, 0)

        self.mask = None
        if hasattr(T, 'get_mask_nyquist'):
            self.mask = T.get_mask_nyquist()

    @classmethod
    def steps(cls):
        raise NotImplementedError

    def stages(self):
        raise NotImplementedError

    def assemble(self):
        a = self.stages()[0]
        dt = self.dt
        ul = copy.copy(self.u)
        ul._basis = TrialFunction(self.u.function_space())
        mats = inner(self.v, ul - dt*a[1, 1]*self.L(ul))
        self.solvers.append(self._solver(mats))
        self.linear_rhs = Inner(self.v, self.L(self.u))
        self.u0_rhs = Inner(self.v, self.u)
        if isinstance(self.N, (Expr, Function)):
            self.nonlinear_rhs = Inner(self.v, self.N)
        elif isinstance(self.N, list):
            self.nonlinear_rhs = sum([Inner(self.v, f) for f in self.N[1:]], start=Inner(self.v, self.N[0]))
        else:
            raise RuntimeError('Wrong type of nonlinear expression')

    def compute_rhs(self, rk=0):
        a, b = self.stages()[:2]
        self.Krhs[rk] = self.nonlinear_rhs()
        if rk == 0:
            self.u0_rhs() # only at start
        self.rhs[:] = self.u0_rhs.output_array
        for j in range(0, rk+1):
            self.rhs += self.dt*b[rk+1, j]*self.Krhs[j]
        if rk > 0:
            self.Lrhs[rk-1] = self.linear_rhs()
            for j in range(0, rk):
                self.rhs += self.dt*a[rk+1, j+1]*self.Lrhs[j]

        if self.mask is not None:
            self.T.mask_nyquist(self.rhs, self.mask)
        return self.rhs

    def solve_step(self, rk=0):
        # only one solver since the diagonal of a is constant
        return self.solvers[0](self.rhs, self.u_)

class IMEXRK111(PDEIMEXRK):

    @classmethod
    def steps(cls):
        return 1

    def stages(self):
        a = np.array([
            [0, 0],
            [0, 1]])
        b = np.array([
            [0, 0],
            [1, 0]])
        c = (0, 1)
        return a, b, c

class IMEXRK222(PDEIMEXRK):

    @classmethod
    def steps(cls):
        return 2

    def stages(self):
        gamma = (2-np.sqrt(2))/2
        delta = 1-1/(2*gamma)
        a = np.array([
            [0, 0, 0],
            [0, gamma, 0],
            [0, 1-gamma, gamma]])
        b = np.array([
            [0, 0, 0],
            [gamma, 0, 0],
            [delta, 1-delta, 0]])
        c = (0, gamma, 1)
        return a, b, c

class IMEXRK443(PDEIMEXRK):

    @classmethod
    def steps(cls):
        return 4

    def stages(self):
        a = np.array([
            [0, 0, 0, 0, 0],
            [0, 1/2, 0, 0, 0],
            [0, 1/6, 1/2, 0, 0],
            [0, -1/2, 1/2, 1/2, 0],
            [0, 3/2, -3/2, 1/2, 1/2]])
        b = np.array([
            [0, 0, 0, 0, 0],
            [1/2, 0, 0, 0, 0],
            [11/18, 1/18, 0, 0, 0],
            [5/6, -5/6, 1/2, 0, 0],
            [1/4, 7/4, 3/4, -7/4, 0]])
        c = (0, 1/2, 2/3, 1/2, 1)
        return a, b, c



import pytest
import numpy as np
import sympy as sp
from shenfun import chebyshev, chebyshevu, legendre, fourier, hermite, \
    laguerre, jacobi, ultraspherical
import importlib

trialbases = []
testbases = []
bcbases = []
for family in ('chebyshev', 'chebyshevu', 'legendre', 'ultraspherical', 'jacobi', 'laguerre', 'hermite', 'fourier'):
    base = importlib.import_module('.'.join(('shenfun', family.lower())))
    trialbases += [base.bases.__dict__.get(b) for b in base.bases.bases if b != 'Generic']
    testbases += [base.bases.__dict__.get(b) for b in base.bases.testbases if b != 'Phi6']
    bcbases += [base.bases.__dict__.get(b) for b in base.bases.bcbases]

nonBC = (
    'Apply',
    'Periodic',
    'Biharmonic*2'
)

@pytest.mark.parametrize('base', trialbases+testbases)
def test_eval_basis(base):
    N = 8
    B = base(N)
    x = sp.symbols('x', real=True)
    b = [B]
    try:
        b += [B.get_bc_space()]
    except:
        pass

    for basis in b:
        M = basis.dim()
        for i in range(M):
            s = basis.basis_function(i, x)
            mesh = np.random.rand(3)
            f0 = sp.lambdify(x, s, 'numpy')(mesh)
            f1 = basis.evaluate_basis(mesh, i=i)
            assert np.allclose(f0, f1)


@pytest.mark.parametrize('base', trialbases+testbases)
def test_eval_basis_derivative(base):
    N = 8
    B = base(N)
    b = [B]
    try:
        b += [B.get_bc_space()]
    except:
        pass
    for basis in b:
        M = basis.dim()
        for i in range(M):
            x = sp.symbols('x', real=True)
            s = basis.basis_function(i, x)
            mesh = np.random.rand(3)
            for k in (1, 2, 3):
                f0 = sp.lambdify(x, s.diff(x, k), 'numpy')(mesh)
                f1 = basis.evaluate_basis_derivative(mesh, i=i, k=k)
                assert np.allclose(f0, f1)

if __name__ == '__main__':
    #test_eval_basis_derivative(chebyshev.Heinrichs)
    test_eval_basis(chebyshev.UpperDirichlet)
